"""Tests classes attribute of classifiers."""

from __future__ import annotations

from typing import Any

import numpy as np
import pytest

from skfda._utils._sklearn_adapter import ClassifierMixin
from skfda.datasets import make_gaussian_process
from skfda.exploratory.depth import ModifiedBandDepth
from skfda.exploratory.stats.covariance import ParametricGaussianCovariance
from skfda.misc.covariances import Gaussian
from skfda.ml.classification import (
    DDClassifier,
    DDGClassifier,
    DTMClassifier,
    KNeighborsClassifier,
    LogisticRegression,
    MaximumDepthClassifier,
    NearestCentroid,
    QuadraticDiscriminantAnalysis,
    RadiusNeighborsClassifier,
)
from skfda.representation import FData

from ..typing._numpy import NDArrayAny


@pytest.fixture(
    params=[
        DDClassifier(degree=2),
        DDGClassifier(
            depth_method=[("mbd", ModifiedBandDepth())],
            multivariate_classifier=KNeighborsClassifier(),
        ),
        DTMClassifier(proportiontocut=0.25),
        KNeighborsClassifier(),
        LogisticRegression(),
        MaximumDepthClassifier(),
        NearestCentroid(),
        QuadraticDiscriminantAnalysis(
            cov_estimator=ParametricGaussianCovariance(
                Gaussian(),
            ),
        ),
        RadiusNeighborsClassifier(),
    ],
    ids=lambda clf: type(clf).__name__,
)
def classifier(request: Any) -> Any:
    """Fixture for classifiers to test."""
    return request.param


@pytest.fixture(
    params=[
        np.array([0, 1]),
        np.array(["class_a", "class_b"]),
    ],
    ids=["int", "str"],
)
def classes(request: Any) -> Any:
    """Fixture for classes to test."""
    return request.param


def test_classes(
    classifier: ClassifierMixin[FData, NDArrayAny],
    classes: NDArrayAny,
) -> None:
    """Test classes attribute of classifiers."""
    n_samples = 30
    y = np.resize(classes, n_samples)
    X = make_gaussian_process(n_samples=n_samples, random_state=0)
    classifier.fit(X, y)
    resulting_classes = np.unique(classifier.predict(X))

    np.testing.assert_array_equal(classifier.classes_, classes)
    np.testing.assert_array_equal(classifier.classes_, resulting_classes)



"""Typing for norms and metrics."""
import enum
from builtins import isinstance
from typing import Any, TypeVar, Union, overload

from typing_extensions import Final, Literal

from ...typing._metric import Metric


class _MetricSingletons(enum.Enum):
    PRECOMPUTED = "precomputed"


PRECOMPUTED: Final = _MetricSingletons.PRECOMPUTED

_PrecomputedTypes = Literal[
    _MetricSingletons.PRECOMPUTED,
    "precomputed",
]

_NonStringMetric = TypeVar(
    "_NonStringMetric",
    bound=Union[
        Metric[Any],
        _MetricSingletons,
    ],
)


@overload
def _parse_metric(
    metric: str,
) -> _MetricSingletons:
    pass


@overload
def _parse_metric(
    metric: _NonStringMetric,
) -> _NonStringMetric:
    pass


def _parse_metric(
    metric: Union[Metric[Any], _MetricSingletons, str],
) -> Union[Metric[Any], _MetricSingletons]:

    return _MetricSingletons(metric) if isinstance(metric, str) else metric



"""Operators applicable to functional data."""
from typing import TYPE_CHECKING

import lazy_loader as lazy

__getattr__, __dir__, __all__ = lazy.attach(
    __name__,
    submod_attrs={
        "_identity": ["Identity"],
        "_integral_transform": ["IntegralTransform"],
        "_linear_differential_operator": ["LinearDifferentialOperator"],
        "_operators": [
            "MatrixOperator",
            "Operator",
            "gram_matrix",
            "gram_matrix_optimization",
        ],
        "_srvf": ["SRSF"],
    },
)

if TYPE_CHECKING:
    from ._identity import Identity as Identity
    from ._integral_transform import IntegralTransform as IntegralTransform
    from ._linear_differential_operator import (
        LinearDifferentialOperator as LinearDifferentialOperator
    )
    from ._operators import (
        MatrixOperator as MatrixOperator,
        Operator as Operator,
        gram_matrix as gram_matrix,
        gram_matrix_optimization as gram_matrix_optimization,
    )
    from ._srvf import SRSF as SRSF



import abc
import warnings
from typing import Callable, Optional

from ...misc import kernels
from ...misc.hat_matrix import (
    HatMatrix,
    KNeighborsHatMatrix,
    LocalLinearRegressionHatMatrix,
    NadarayaWatsonHatMatrix,
)
from ...typing._base import GridPointsLike
from ...typing._numpy import NDArrayFloat
from . import KernelSmoother
from ._linear import _LinearSmoother

warnings.warn(
    "The \"kernel_smoothers\" module is deprecated. "
    "Use the \"KernelSmoother\" class instead",
    DeprecationWarning,
    stacklevel=2,
)


class _DeprecatedLinearKernelSmoother(_LinearSmoother):

    def __init__(
        self,
        *,
        smoothing_parameter: Optional[float] = None,
        kernel: Callable[[NDArrayFloat], NDArrayFloat] = kernels.normal,
        weights: Optional[NDArrayFloat] = None,
        output_points: Optional[GridPointsLike] = None,
    ):
        self.smoothing_parameter = smoothing_parameter
        self.kernel = kernel
        self.weights = weights
        self.output_points = output_points

        warnings.warn(
            f"Class \"{type(self)}\" is deprecated. "
            "Use the \"KernelSmoother\" class instead",
            DeprecationWarning,
        )

    def _hat_matrix(
        self,
        input_points: GridPointsLike,
        output_points: GridPointsLike,
    ) -> NDArrayFloat:

        return KernelSmoother(
            kernel_estimator=self._get_kernel_estimator(),
            weights=self.weights,
            output_points=output_points,
        )._hat_matrix(
            input_points=input_points,
            output_points=output_points,
        )

    @abc.abstractmethod
    def _get_kernel_estimator(self) -> HatMatrix:
        raise NotImplementedError


class NadarayaWatsonSmoother(_DeprecatedLinearKernelSmoother):
    """Nadaraya-Watson smoother (deprecated)."""

    def _get_kernel_estimator(self) -> HatMatrix:
        return NadarayaWatsonHatMatrix(
            bandwidth=self.smoothing_parameter,
            kernel=self.kernel,
        )


class LocalLinearRegressionSmoother(_DeprecatedLinearKernelSmoother):
    """Local linear regression smoother (deprecated)."""

    def _get_kernel_estimator(self) -> HatMatrix:
        return LocalLinearRegressionHatMatrix(
            bandwidth=self.smoothing_parameter,
            kernel=self.kernel,
        )


class KNeighborsSmoother(_DeprecatedLinearKernelSmoother):
    """Local linear regression smoother (deprecated)."""

    def __init__(
        self,
        *,
        smoothing_parameter: Optional[int] = None,
        kernel: Callable[[NDArrayFloat], NDArrayFloat] = kernels.uniform,
        weights: Optional[NDArrayFloat] = None,
        output_points: Optional[GridPointsLike] = None,
    ):
        super().__init__(
            smoothing_parameter=smoothing_parameter,
            kernel=kernel,
            weights=weights,
            output_points=output_points,
        )

    def _get_kernel_estimator(self) -> HatMatrix:
        return KNeighborsHatMatrix(
            n_neighbors=(
                int(self.smoothing_parameter)
                if self.smoothing_parameter is not None
                else None
            ),
            kernel=self.kernel,
        )



from __future__ import annotations

import numbers
from typing import Callable, Sequence, Tuple, Union

import findiff
import numpy as np
import scipy.integrate
from numpy import polyder, polyint, polymul, polyval
from scipy.interpolate import PPoly

from ...representation import FData, FDataBasis, FDataGrid
from ...representation.basis import (
    Basis,
    BSplineBasis,
    ConstantBasis,
    CustomBasis,
    FourierBasis,
    MonomialBasis,
    _GridBasis,
)
from ...typing._base import DomainRangeLike
from ...typing._numpy import NDArrayFloat, NDArrayInt
from ._operators import Operator, gram_matrix, gram_matrix_optimization

Order = int

WeightSequence = Sequence[
    Union[
        float,
        Callable[[NDArrayFloat], NDArrayFloat],
    ],
]


class LinearDifferentialOperator(
    Operator[Union[FData, Basis], Callable[[NDArrayFloat], NDArrayFloat]],
):
    r"""
    Defines the structure of a linear differential operator function system.

    .. math::
        Lx(t) = b_0(t) x(t) + b_1(t) x'(t) +
                \dots + b_{n-1}(t) D^{n-1}(x(t)) + b_n(t) D^n(x(t))

    Can only be applied to functional data, as multivariate data has no
    derivatives.

    You have to provide either order or weights.
    If both are provided, it will raise an error.
    If a positional argument is supplied it will be considered the
    order if it is an integral type and the weights otherwise.

    Parameters:
        order: the order of the operator. It's the highest derivative order
            of the operator.
        weights: A FDataBasis objects list of length order + 1 items.
        domain_range: Definition of the interval where the weight functions
            are defined. If the functional weights are specified and this is
            not, takes the domain range from them. Otherwise, defaults to
            (0,1).

    Attributes:
        weights (list):  A list of callables.

    Examples:
        Create a linear differential operator that penalizes the second
        derivative (acceleration)

        >>> from skfda.misc.operators import LinearDifferentialOperator
        >>> from skfda.representation.basis import (FDataBasis,
        ...                                         MonomialBasis,
        ...                                         ConstantBasis)
        >>>
        >>> LinearDifferentialOperator(2)
        LinearDifferentialOperator(
            weights=(0, 0, 1),
        )

        Create a linear differential operator that penalizes three times
        the second derivative (acceleration) and twice the first (velocity).

        >>> LinearDifferentialOperator(weights=[0, 2, 3])
        LinearDifferentialOperator(
            weights=(0, 2, 3),
        )

        Create a linear differential operator with non-constant weights.

        >>> constant = ConstantBasis()
        >>> monomial = MonomialBasis(domain_range=(0, 1), n_basis=3)
        >>> fdlist = [FDataBasis(constant, [0.]),
        ...           FDataBasis(constant, [0.]),
        ...           FDataBasis(monomial, [1., 2., 3.])]
        >>> LinearDifferentialOperator(weights=fdlist)
        LinearDifferentialOperator(
        weights=(FDataBasis(
                basis=ConstantBasis(domain_range=((0.0, 1.0),), n_basis=1),
                coefficients=[[ 0.]],
                ...),
            FDataBasis(
                basis=ConstantBasis(domain_range=((0.0, 1.0),), n_basis=1),
                coefficients=[[ 0.]],
                ...),
            FDataBasis(
                basis=MonomialBasis(domain_range=((0.0, 1.0),), n_basis=3),
                coefficients=[[ 1. 2. 3.]],
                ...)),
        )

    """

    def __init__(
        self,
        order_or_weights: Order | WeightSequence | None = None,
        *,
        order: int | None = None,
        weights: WeightSequence | None = None,
        domain_range: DomainRangeLike | None = None,
    ) -> None:

        num_args = sum(
            a is not None for a in (order_or_weights, order, weights)
        )

        if num_args > 1:
            raise ValueError(
                "You have to provide the order or the weights, not both.",
            )

        if order_or_weights is not None:
            if isinstance(order_or_weights, numbers.Integral):
                order = int(order_or_weights)
            else:
                assert isinstance(order_or_weights, Sequence)
                weights = order_or_weights

        if order is None and weights is None:
            weights = (0,)

        elif order is not None:
            if order < 0:
                raise ValueError("Order should be an non-negative integer.")

            weights = tuple(
                0 if (i < order) else 1
                for i in range(order + 1)
            )

        assert weights is not None
        if len(weights) == 0:
            raise ValueError("You have to provide one weight at least.")

        # Check domain ranges
        for w in weights:
            w_domain_range = getattr(w, "domain_range", None)

            if w_domain_range is not None:
                if domain_range is None:
                    domain_range = w_domain_range
                elif not np.array_equal(w_domain_range, domain_range):
                    raise ValueError(
                        "Weights with wrong domain range.",
                    )
        self.weights = tuple(weights)

    def __repr__(self) -> str:

        return (
            f"{self.__class__.__name__}(\n"
            f"\tweights={self.weights},\n"
            f")"
        ).replace('\n', '\n    ')

    def __eq__(self, other: object) -> bool:

        return (
            isinstance(other, LinearDifferentialOperator)
            and self.weights == other.weights
        )

    def constant_weights(self) -> NDArrayFloat | None:
        """
        Return constant weights.

        Return the scalar weights of the linear differential operator if they
        are constant basis.
        Otherwise, return None.

        This function is mostly useful for basis which want to override
        the _penalty method in order to use an analytical expression
        for constant weights.

        """
        weights = np.array(self.weights)

        return None if weights.dtype == np.object_ else weights

    def __call__(
        self,
        f: FData | Basis,
    ) -> Callable[[NDArrayFloat], NDArrayFloat]:
        """Return the function that results of applying the operator."""
        function_derivatives = [
            f.derivative(order=i) for i, _ in enumerate(self.weights)
        ]

        def applied_linear_diff_op(
            t: NDArrayFloat,
        ) -> NDArrayFloat:
            result = sum(
                (w(t) if callable(w) else np.asarray(w))
                * function_derivatives[i](t)
                for i, w in enumerate(self.weights)
            )
            assert not isinstance(result, int)
            return result

        return applied_linear_diff_op


#############################################################
#
# Optimized implementations of gram matrix for each basis.
#
#############################################################


@gram_matrix_optimization.register
def constant_penalty_matrix_optimized(
    linear_operator: LinearDifferentialOperator,
    basis: ConstantBasis,
) -> NDArrayFloat:
    """Optimized version for Constant basis."""
    coefs = linear_operator.constant_weights()
    if coefs is None:
        return NotImplemented

    return np.array([[
        coefs[0] ** 2
        * (basis.domain_range[0][1] - basis.domain_range[0][0]),
    ]])


def _monomial_evaluate_constant_linear_diff_op(
    basis: MonomialBasis,
    weights: NDArrayFloat,
) -> NDArrayFloat:
    """Evaluate constant weights over the monomial basis."""
    max_derivative = len(weights) - 1

    seq = np.arange(basis.n_basis)
    coef_mat = np.linspace(
        seq,
        seq - max_derivative + 1,
        max_derivative,
        dtype=int,
    )

    # Compute coefficients for each derivative
    coefs = np.cumprod(coef_mat, axis=0)

    # Add derivative 0 row
    coefs = np.concatenate((np.ones((1, basis.n_basis)), coefs))

    # Now each row correspond to each basis and each column to
    # each derivative
    coefs_t = coefs.T

    # Multiply by the weights
    weighted_coefs = coefs_t * weights
    assert len(weighted_coefs) == basis.n_basis

    # Now each row has the right weight, but the polynomials are in a
    # decreasing order and with different exponents

    # Resize the coefs so that there are as many rows as the number of
    # basis
    # The matrix is now triangular
    # refcheck is False to prevent exceptions while debugging
    weighted_coefs = np.copy(weighted_coefs.T)
    weighted_coefs.resize(
        basis.n_basis,
        basis.n_basis,
        refcheck=False,
    )
    weighted_coefs = weighted_coefs.T

    # Shift the coefficients so that they correspond to the right
    # exponent
    indexes = np.tril_indices(basis.n_basis)
    polynomials = np.zeros_like(weighted_coefs)
    polynomials[
        indexes[0],
        indexes[1] - indexes[0] - 1,
    ] = weighted_coefs[indexes]

    # At this point, each row of the matrix correspond to a polynomial
    # that is the result of applying the linear differential operator
    # to each element of the basis

    return polynomials  # type: ignore[no-any-return]


@gram_matrix_optimization.register
def monomial_penalty_matrix_optimized(
    linear_operator: LinearDifferentialOperator,
    basis: MonomialBasis,
) -> NDArrayFloat:
    """Optimized version for Monomial basis."""
    weights = linear_operator.constant_weights()
    if weights is None:
        return NotImplemented

    polynomials = _monomial_evaluate_constant_linear_diff_op(basis, weights)

    # Expand the polinomials with 0, so that the multiplication fits
    # inside. It will need the double of the degree
    length_with_padding = polynomials.shape[1] * 2 - 1

    # Multiplication of polynomials is a convolution.
    # The convolution can be performed in parallel applying a Fourier
    # transform and then doing a normal multiplication in that
    # space, coverting back with the inverse Fourier transform
    fft = np.fft.rfft(polynomials, length_with_padding)

    # We compute only the upper matrix, as the penalty matrix is
    # symmetrical
    indices = np.triu_indices(basis.n_basis)
    fft_mul = fft[indices[0]] * fft[indices[1]]

    integrand = np.fft.irfft(fft_mul, length_with_padding)

    integration_domain = basis.domain_range[0]

    # To integrate, divide by the position and increase the exponent
    # in the evaluation
    denom = np.arange(integrand.shape[1], 0, -1)
    integrand /= denom

    # Add column of zeros at the right to increase exponent
    integrand = np.pad(
        integrand,
        pad_width=(
            (0, 0),
            (0, 1),
        ),
        mode='constant',
    )

    # Now, apply Barrow's rule
    # polyval applies Horner method over the first dimension,
    # so we need to transpose
    x_right = np.polyval(integrand.T, integration_domain[1])
    x_left = np.polyval(integrand.T, integration_domain[0])

    integral = x_right - x_left

    penalty_matrix = np.empty((basis.n_basis, basis.n_basis))

    # Set upper matrix
    penalty_matrix[indices] = integral

    # Set lower matrix
    penalty_matrix[(indices[1], indices[0])] = integral

    return penalty_matrix


def _fourier_penalty_matrix_optimized_orthonormal(
    basis: FourierBasis,
    weights: NDArrayFloat,
) -> NDArrayFloat:
    """Return the penalty when the basis is orthonormal."""
    signs = np.array([1, 1, -1, -1])
    signs_expanded = np.tile(signs, len(weights) // 4 + 1)

    signs_odd = signs_expanded[:len(weights)]
    signs_even = signs_expanded[1:len(weights) + 1]

    phases = (
        np.arange(1, (basis.n_basis - 1) // 2 + 1)
        * 2 * np.pi / basis.period
    )

    # Compute increasing powers
    coefs_no_sign = np.vander(phases, len(weights), increasing=True)

    coefs_no_sign *= weights

    coefs_odd = signs_odd * coefs_no_sign
    coefs_even = signs_even * coefs_no_sign

    # After applying the linear differential operator to a sinusoidal
    # element of the basis e, the result can be expressed as
    # A e + B e*, where e* is the other basis element in the pair
    # with the same phase

    odd_sin_coefs = np.sum(coefs_odd[:, ::2], axis=1)
    odd_cos_coefs = np.sum(coefs_odd[:, 1::2], axis=1)

    even_cos_coefs = np.sum(coefs_even[:, ::2], axis=1)
    even_sin_coefs = np.sum(coefs_even[:, 1::2], axis=1)

    # The diagonal is the inner product of A e + B e*
    # with itself. As the basis is orthonormal, the cross products e e*
    # are 0, and the products e e and e* e* are one.
    # Thus, the diagonal is A^2 + B^2
    # All elements outside the main diagonal are 0
    main_diag_odd = odd_sin_coefs**2 + odd_cos_coefs**2
    main_diag_even = even_sin_coefs**2 + even_cos_coefs**2

    # The main diagonal should intercalate both diagonals
    main_diag = np.array((main_diag_odd, main_diag_even)).T.ravel()

    penalty_matrix = np.diag(main_diag)

    # Add row and column for the constant
    penalty_matrix = np.pad(
        penalty_matrix,
        pad_width=(
            (1, 0),
            (1, 0),
        ),
        mode='constant',
    )

    penalty_matrix[0, 0] = weights[0]**2

    return penalty_matrix


@gram_matrix_optimization.register
def fourier_penalty_matrix_optimized(
    linear_operator: LinearDifferentialOperator,
    basis: FourierBasis,
) -> NDArrayFloat:
    """Optimized version for Fourier basis."""
    weights = linear_operator.constant_weights()
    if weights is None:
        return NotImplemented

    # If the period and domain range are not the same, the basis functions
    # are not orthogonal
    if basis.period != (basis.domain_range[0][1] - basis.domain_range[0][0]):
        return NotImplemented

    return _fourier_penalty_matrix_optimized_orthonormal(basis, weights)


@gram_matrix_optimization.register
def bspline_penalty_matrix_optimized(
    linear_operator: LinearDifferentialOperator,
    basis: BSplineBasis,
) -> NDArrayFloat:
    """Optimized version for BSpline basis."""
    coefs = linear_operator.constant_weights()
    if coefs is None:
        return NotImplemented

    nonzero = np.flatnonzero(coefs)

    # All derivatives above the order of the spline are effectively
    # zero
    nonzero = nonzero[nonzero < basis.order]

    if len(nonzero) == 0:
        return np.zeros((basis.n_basis, basis.n_basis))

    # We will only deal with one nonzero coefficient right now
    if len(nonzero) != 1:
        return NotImplemented

    derivative_degree = nonzero[0]

    if derivative_degree == basis.order - 1:
        # The derivative of the bsplines are constant in the intervals
        # defined between knots
        knots = np.array(basis.knots)
        mid_inter = (knots[1:] + knots[:-1]) / 2
        basis_deriv = basis.derivative(order=derivative_degree)
        constants = basis_deriv(mid_inter)[..., 0].T
        knots_intervals = np.diff(basis.knots)
        # Integration of product of constants
        return (  # type: ignore[no-any-return]
            constants.T @ np.diag(knots_intervals) @ constants
        )

    # We only deal with the case without zero length intervals
    # for now
    if np.any(np.diff(basis.knots) == 0):
        return NotImplemented

    # Compute exactly using the piecewise polynomial
    # representation of splines

    # Places m knots at the boundaries
    knots = np.array(basis._evaluation_knots())

    # c is used the select which spline the function
    # PPoly.from_spline below computes
    c = np.zeros(len(knots))

    # Initialise empty list to store the piecewise polynomials
    ppoly_lst = []

    no_0_intervals = np.where(np.diff(knots) > 0)[0]

    # For each basis gets its piecewise polynomial representation
    for i in range(basis.n_basis):

        # Write a 1 in c in the position of the spline
        # transformed in each iteration
        c[i] = 1

        # Gets the piecewise polynomial representation and gets
        # only the positions for no zero length intervals
        # This polynomial are defined relatively to the knots
        # meaning that the column i corresponds to the ith knot.
        # Let the ith knot be a
        # Then f(x) = pp(x - a)
        pp = PPoly.from_spline((knots, c, basis.order - 1))
        pp_coefs = pp.c[:, no_0_intervals]

        # We have the coefficients for each interval in coordinates
        # (x - a), so we will need to subtract a when computing the
        # definite integral
        ppoly_lst.append(pp_coefs)
        c[i] = 0

    # Now for each pair of basis computes the inner product after
    # applying the linear differential operator
    penalty_matrix = np.zeros((basis.n_basis, basis.n_basis))
    for interval, _ in enumerate(no_0_intervals):
        for i in range(basis.n_basis):
            poly_i = np.trim_zeros(ppoly_lst[i][:, interval], 'f')
            if len(poly_i) <= derivative_degree:
                # if the order of the polynomial is lesser or
                # equal to the derivative the result of the
                # integral will be 0
                continue
            # indefinite integral
            derivative = polyder(poly_i, derivative_degree)
            square = polymul(derivative, derivative)
            integral = polyint(square)

            # definite integral
            penalty_matrix[i, i] += np.diff(
                polyval(
                    integral,
                    np.asarray(basis.knots[interval: interval + 2])
                    - basis.knots[interval],
                ),
            )[0]

            for j in range(i + 1, basis.n_basis):
                poly_j = np.trim_zeros(ppoly_lst[j][:, interval], 'f')
                if len(poly_j) <= derivative_degree:
                    # if the order of the polynomial is lesser
                    # or equal to the derivative the result of
                    # the integral will be 0
                    continue
                    # indefinite integral
                integral = polyint(
                    polymul(
                        polyder(poly_i, derivative_degree),
                        polyder(poly_j, derivative_degree),
                    ),
                )
                # definite integral
                penalty_matrix[i, j] += np.diff(
                    polyval(
                        integral,
                        np.asarray(basis.knots[interval: interval + 2])
                        - basis.knots[interval],
                    ),
                )[0]
                penalty_matrix[j, i] = penalty_matrix[i, j]
    return penalty_matrix


def _optimized_operator_evaluation_in_grid(
    linear_operator: LinearDifferentialOperator,
    grid_points: NDArrayFloat,
    findiff_accuracy: int,
) -> Tuple[NDArrayFloat, NDArrayInt]:
    """
    Compute the linear operator applied to the delta basis of the grid.

    The delta basis is the basis where the i-th basis function is 1 in the
    i-th point and 0 in the rest of points.

    Returns:
    If the linear operator does not have constant weights, NotImplemented.
    Otherwise, returns a tuple with:
     - A matrix where the i-th row contains the values of the linear operator
        applied to the i-th basis function of the delta basis.

    - The domain limits of each basis function of the delta basis. Outside
        these limits, the result of applying the linear operator to the
        given basis function is zero.

    Example:
        >>> import numpy as np
        >>> from skfda.misc.operators import LinearDifferentialOperator
        >>> _optimized_operator_evaluation_in_grid(
        ...     LinearDifferentialOperator(2),
        ...     np.linspace(0, 7, 8),
        ...     2,
        ... )
        (array([[ 2.,  1.,  0.,  0.,  0.,  0.,  0.,  0.],
        [-5., -2.,  1.,  0.,  0.,  0.,  0.,  0.],
        [ 4.,  1., -2.,  1.,  0.,  0.,  0.,  0.],
        [-1.,  0.,  1., -2.,  1.,  0.,  0.,  0.],
        [ 0.,  0.,  0.,  1., -2.,  1.,  0., -1.],
        [ 0.,  0.,  0.,  0.,  1., -2.,  1.,  4.],
        [ 0.,  0.,  0.,  0.,  0.,  1., -2., -5.],
        [ 0.,  0.,  0.,  0.,  0.,  0.,  1.,  2.]]),
        array([[0, 1],
                [0, 2],
                [0, 3],
                [0, 4],
                [3, 7],
                [4, 7],
                [5, 7],
                [6, 7]]))

    Explanation:
        Each row of the first array contains the values of the linear operator
        applied to the delta basis function.
        As it can be seen, each row of the second array cointains the domain
        limits of the corresponding row of the first array. That is to say,
        the range in which their values are not 0.
        For example, the third row of the first array is non-zero until the
        fourth element, so the third row of the second array is [0, 3].
    """
    n_points = grid_points.shape[0]
    diff_coefficients = np.zeros((n_points, n_points))

    domain_limits = np.hstack(
        [
            np.ones((n_points, 1)) * n_points,
            np.zeros((n_points, 1)),
        ],
    ).astype(int)

    linear_operator_weights = linear_operator.constant_weights()
    if linear_operator_weights is None:
        return NotImplemented, NotImplemented

    # The desired matrix can be obtained by appending the coefficients
    # of the finite differences for each point column-wise.
    for dif_order, w in enumerate(linear_operator_weights):
        if w == 0:
            continue
        for point_index in range(n_points):
            # Calculate the finite differences coefficients
            # the point point_index. These coefficients express the
            # contribution of a function in each of the points in the
            # grid to the value of the operator applied to the function
            # in the point point_index.
            finite_diff = findiff.coefs.coefficients_non_uni(
                deriv=dif_order,
                acc=findiff_accuracy,
                coords=grid_points,
                idx=point_index,
            )

            # The offsets are the relative positions of the
            # points for which the coefficients are being
            # calculated
            offsets = finite_diff["offsets"]
            coefficients = finite_diff["coefficients"]

            # Add the position of the point where the finite
            # difference is calculated to get the absolute position
            # of the points where each coefficient is applied
            positions = offsets + point_index

            # If the coefficients for calculating the value of
            # the operator in this point (point_index)
            # include a coefficient for the i-th point,
            # then the i-th basis function contains the point point_index
            # in its domain.
            # (by definition that fuction is 1 in the i-th point).
            domain_limits[positions, 0] = np.minimum(
                domain_limits[positions, 0],
                np.ones(positions.shape) * point_index,
            )
            domain_limits[positions, 1] = np.maximum(
                domain_limits[positions, 1],
                np.ones(positions.shape) * point_index,
            )

            w_eval = w(grid_points[point_index]) if callable(w) else w
            diff_coefficients[positions, point_index] += w_eval * coefficients

    return (diff_coefficients, domain_limits)


@gram_matrix_optimization.register
def gridbasis_penalty_matrix_optimized(
    linear_operator: LinearDifferentialOperator,
    basis: _GridBasis,
) -> NDArrayFloat:
    """
    Optimized version for GridBasis.

    It evaluates the operator in the grid points.
    """
    n_points = basis.grid_points[0].shape[0]
    findiff_accuracy = 2

    evaluated_basis, domain_limits = _optimized_operator_evaluation_in_grid(
        linear_operator,
        basis.grid_points[0],
        findiff_accuracy=findiff_accuracy,
    )

    # If the operator does not have constant weights, the
    # optimized evaluation is not implemented.
    if evaluated_basis is NotImplemented:
        return NotImplemented

    # The support of the result of applying the linear operator
    # is a small subset of the grid points. We calculate the
    # maximum number of points to the left and right of the
    # point where the linear operator is applied that can be
    # different from zero.

    max_domain_width = np.max(
        np.abs(domain_limits[:, 0] - domain_limits[:, 1]),
    )

    # Precompute the simpson quadrature for the entire domain.
    # This is done to avoid computing it for each subinterval of integration.
    quadrature = scipy.integrate.simpson(
        np.identity(n_points),
        basis.grid_points[0],
    )

    product_matrix = np.zeros((basis.n_basis, basis.n_basis))
    for i in range(n_points):
        # Maximum index of another function with not-disjoint support
        max_no_disjoint_index = min(n_points, i + 2 * max_domain_width)

        for j in range(i, max_no_disjoint_index):
            if domain_limits[i, 1] < domain_limits[j, 0]:
                continue

            left_lim = min(
                domain_limits[i, 0],
                domain_limits[j, 0],
            )
            right_lim = max(
                domain_limits[i, 1],
                domain_limits[j, 1],
            )

            # The right limit of the interval is excluded when slicing
            right_lim += 1

            # Limit both the multiplication and the integration to
            # the interval where the functions are not 0.
            operator_product = (
                evaluated_basis[i, left_lim:right_lim]
                * evaluated_basis[j, left_lim:right_lim]
            )
            sub_quadrature = quadrature[left_lim:right_lim]

            product_matrix[i, j] = np.dot(operator_product, sub_quadrature)
            product_matrix[j, i] = product_matrix[i, j]

    return product_matrix


@gram_matrix_optimization.register
def fdatabasis_penalty_matrix_optimized(
    linear_operator: LinearDifferentialOperator,
    fdatabasis: FDataBasis,
) -> NDArrayFloat:
    """Optimized version for FDataBasis."""
    # By calculating the gram matrix of the basis first, we can
    # take advantage of the optimized implementations for
    # several basis types.
    # Then we can calculate the penalty matrix using the
    # coefficients of the functions in the basis
    # and the penalty matrix of the basis.
    basis_pen_matrix = gram_matrix(
        linear_operator,
        fdatabasis.basis,
    )

    coef_matrix = fdatabasis.coefficients
    return coef_matrix @ basis_pen_matrix @ coef_matrix.T


@gram_matrix_optimization.register
def fdatagrid_penalty_matrix_optimized(
    linear_operator: LinearDifferentialOperator,
    fdatagrid: FDataGrid,
) -> NDArrayFloat:
    """Optimized version for FDataGrid."""
    from ...misc import inner_product_matrix

    operator_evaluated = linear_operator(fdatagrid)(
        fdatagrid.grid_points[0],
    )
    fdatagrid_evaluated = FDataGrid(
        data_matrix=operator_evaluated[..., 0],
        grid_points=fdatagrid.grid_points[0],
    )
    return inner_product_matrix(fdatagrid_evaluated)


@gram_matrix_optimization.register
def custombasis_penalty_matrix_optimized(
    linear_operator: LinearDifferentialOperator,
    basis: CustomBasis,
) -> NDArrayFloat:
    """Optimized version for CustomBasis."""
    return gram_matrix(linear_operator, basis.fdata)



"""Tests for ANOVA."""
import unittest

import numpy as np

from skfda.datasets import fetch_gait
from skfda.inference.anova import (
    oneway_anova,
    v_asymptotic_stat,
    v_sample_stat,
)
from skfda.representation import FDataGrid
from skfda.representation.basis import FourierBasis


class OnewayAnovaTests(unittest.TestCase):
    """Tests for ANOVA."""

    def test_oneway_anova_args(self) -> None:
        """Check behavior of test with invalid args."""
        with self.assertRaises(ValueError):
            oneway_anova(FDataGrid([0]), n_reps=-2)

    def test_v_stats_args(self) -> None:
        """Check behavior of statistic with invalid args."""
        with self.assertRaises(ValueError):
            v_sample_stat(FDataGrid([0]), [0, 1])
        with self.assertRaises(ValueError):
            v_asymptotic_stat(FDataGrid([0]), weights=[0, 1])
        with self.assertRaises(ValueError):
            v_asymptotic_stat(
                FDataGrid([[1, 1, 1], [1, 1, 1]]),
                weights=[0, 0],
            )

    def test_v_stats(self) -> None:
        """Test statistic behaviour."""
        n_features = 50
        weights = [1, 2, 3]
        t = np.linspace(0, 1, n_features)
        m1 = [1 for _ in range(n_features)]
        m2 = [2 for _ in range(n_features)]
        m3 = [3 for _ in range(n_features)]
        fd = FDataGrid([m1, m2, m3], grid_points=t)
        self.assertEqual(v_sample_stat(fd, weights), 7.0)
        self.assertAlmostEqual(
            v_sample_stat(
                fd.to_basis(FourierBasis(n_basis=5)),
                weights,
            ),
            7.0,
        )
        res = (
            (1 - 2 * np.sqrt(1 / 2)) ** 2
            + (1 - 3 * np.sqrt(1 / 3)) ** 2
            + (2 - 3 * np.sqrt(2 / 3)) ** 2
        )
        self.assertAlmostEqual(v_asymptotic_stat(fd, weights=weights), res)
        self.assertAlmostEqual(
            v_asymptotic_stat(
                fd.to_basis(FourierBasis(n_basis=5)),
                weights=weights,
            ),
            res,
        )

    def test_asymptotic_behaviour(self) -> None:
        """Test asymptotic behaviour."""
        dataset = fetch_gait()
        fd = dataset['data'].coordinates[1]
        fd1 = fd[:5]
        fd2 = fd[5:10]
        fd3 = fd[10:15]

        n_little_sim = 10

        sims = np.array([
            oneway_anova(
                fd1,
                fd2,
                fd3,
                n_reps=500,
                random_state=i,
            )[1]
            for i in range(n_little_sim)
        ])
        little_sim = np.mean(sims)
        big_sim = oneway_anova(fd1, fd2, fd3, n_reps=2000, random_state=100)[1]
        self.assertAlmostEqual(little_sim, big_sim, delta=0.05)

        fd = fd.to_basis(FourierBasis(n_basis=5))
        fd1 = fd[:5]
        fd2 = fd[5:10]

        sims = np.array([
            oneway_anova(
                fd1,
                fd2,
                n_reps=500,
                random_state=i,
            )[1]
            for i in range(n_little_sim)
        ])
        little_sim = np.mean(sims)
        big_sim = oneway_anova(fd1, fd2, n_reps=2000, random_state=100)[1]
        self.assertAlmostEqual(little_sim, big_sim, delta=0.05)


if __name__ == '__main__':
    unittest.main()



"""Neighbor models for supervised classification."""

from __future__ import annotations

from typing import Sequence, TypeVar, Union, overload

from sklearn.neighbors import (
    KNeighborsClassifier as _KNeighborsClassifier,
    RadiusNeighborsClassifier as _RadiusNeighborsClassifier,
)
from typing_extensions import Literal

from ..._utils._neighbors_base import (
    AlgorithmType,
    KNeighborsMixin,
    NeighborsClassifierMixin,
    RadiusNeighborsMixin,
    WeightsType,
)
from ...misc.metrics import l2_distance
from ...representation import FData
from ...typing._metric import Metric
from ...typing._numpy import NDArrayFloat, NDArrayInt

InputBound = Union[NDArrayFloat, FData]
Input = TypeVar("Input", contravariant=True, bound=InputBound)
OutlierLabelType = Union[int, str, Sequence[int], Sequence[str], None]


class KNeighborsClassifier(
    KNeighborsMixin[Input, NDArrayInt],
    NeighborsClassifierMixin[Input, NDArrayInt],
):
    """
    Classifier implementing the k-nearest neighbors vote.

    Parameters:
        n_neighbors: Number of neighbors to use by default for
            :meth:`kneighbors` queries.
        weights: Weight function used in prediction.
            Possible values:

            - 'uniform': uniform weights. All points in each neighborhood
              are weighted equally.
            - 'distance': weight points by the inverse of their distance.
              in this case, closer neighbors of a query point will have a
              greater influence than neighbors which are further away.
            - [callable]: a user-defined function which accepts an
              array of distances, and returns an array of the same shape
              containing the weights.

        algorithm: Algorithm used to compute the nearest neighbors:

            - 'ball_tree' will use :class:`sklearn.neighbors.BallTree`.
            - 'brute' will use a brute-force search.
            - 'auto' will attempt to decide the most appropriate algorithm
              based on the values passed to :meth:`fit` method.

        leaf_size: Leaf size passed to BallTree or KDTree. This can affect the
            speed of the construction and query, as well as the memory
            required to store the tree. The optimal value depends on the
            nature of the problem.
        metric: The distance metric to use for the tree. The default metric is
            the L2 distance. See the documentation of the metrics module
            for a list of available metrics.
        n_jobs: The number of parallel jobs to run for neighbors search.
            ``None`` means 1 unless in a :obj:`joblib.parallel_backend`
            context.
            ``-1`` means using all processors.
            Doesn't affect :meth:`fit` method.

    Examples:
        Firstly, we will create a toy dataset with 2 classes

        >>> from skfda.datasets import make_sinusoidal_process
        >>> fd1 = make_sinusoidal_process(phase_std=.25, random_state=0)
        >>> fd2 = make_sinusoidal_process(phase_mean=1.8, error_std=0.,
        ...                               phase_std=.25, random_state=0)
        >>> fd = fd1.concatenate(fd2)
        >>> y = 15*[0] + 15*[1]

        We will fit a K-Nearest Neighbors classifier

        >>> from skfda.ml.classification import KNeighborsClassifier
        >>> neigh = KNeighborsClassifier()
        >>> neigh.fit(fd, y)
        KNeighborsClassifier(...)

        We can predict the class of new samples

        >>> neigh.predict(fd[::2]) # Predict labels for even samples
        array([0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1])

        And the estimated probabilities.

        >>> neigh.predict_proba(fd[0]) # Probabilities of sample 0
        array([[ 1.,  0.]])

    See also:
        :class:`~skfda.ml.classification.RadiusNeighborsClassifier`
        :class:`~skfda.ml.classification.NearestCentroid`
        :class:`~skfda.ml.regression.KNeighborsRegressor`
        :class:`~skfda.ml.regression.RadiusNeighborsRegressor`
        :class:`~skfda.ml.clustering.NearestNeighbors`

    Notes:
        See Nearest Neighbors in the sklearn online documentation for a
        discussion of the choice of ``algorithm`` and ``leaf_size``.

        This class wraps the sklearn classifier
        `sklearn.neighbors.KNeighborsClassifier`.

    Warning:
        Regarding the Nearest Neighbors algorithms, if it is found that two
        neighbors, neighbor `k+1` and `k`, have identical distances
        but different labels, the results will depend on the ordering of the
        training data.

        https://en.wikipedia.org/wiki/K-nearest_neighbor_algorithm
    """

    @overload
    def __init__(
        self: KNeighborsClassifier[NDArrayFloat],
        *,
        n_neighbors: int = 5,
        weights: WeightsType = 'uniform',
        algorithm: AlgorithmType = 'auto',
        leaf_size: int = 30,
        metric: Literal["precomputed"],
        n_jobs: int | None = None,
    ) -> None:
        pass

    @overload
    def __init__(
        self: KNeighborsClassifier[InputBound],
        *,
        n_neighbors: int = 5,
        weights: WeightsType = 'uniform',
        algorithm: AlgorithmType = 'auto',
        leaf_size: int = 30,
        n_jobs: int | None = None,
    ) -> None:
        pass

    @overload
    def __init__(
        self,
        *,
        n_neighbors: int = 5,
        weights: WeightsType = 'uniform',
        algorithm: AlgorithmType = 'auto',
        leaf_size: int = 30,
        metric: Metric[Input] = l2_distance,
        n_jobs: int | None = None,
    ) -> None:
        pass

    # Not useless, it restricts parameters
    def __init__(  # noqa: WPS612
        self,
        *,
        n_neighbors: int = 5,
        weights: WeightsType = 'uniform',
        algorithm: AlgorithmType = 'auto',
        leaf_size: int = 30,
        metric: Literal["precomputed"] | Metric[Input] = l2_distance,
        n_jobs: int | None = None,
    ) -> None:
        super().__init__(
            n_neighbors=n_neighbors,
            weights=weights,
            algorithm=algorithm,
            leaf_size=leaf_size,
            metric=metric,
            n_jobs=n_jobs,
        )

    def _init_estimator(self) -> _KNeighborsClassifier:

        return _KNeighborsClassifier(
            n_neighbors=self.n_neighbors,
            weights=self.weights,
            algorithm=self.algorithm,
            leaf_size=self.leaf_size,
            metric="precomputed",
            n_jobs=self.n_jobs,
        )


class RadiusNeighborsClassifier(
    RadiusNeighborsMixin[Input, NDArrayInt],
    NeighborsClassifierMixin[Input, NDArrayInt],
):
    """
    Classifier implementing a vote among neighbors within a given radius.

    Parameters:
        radius: Range of parameter space to use by default for
            :meth:`radius_neighbors` queries.
        weights: Weight function used in prediction.
            Possible values:

            - 'uniform': uniform weights. All points in each neighborhood
                are weighted equally.
            - 'distance': weight points by the inverse of their distance.
                in this case, closer neighbors of a query point will have a
                greater influence than neighbors which are further away.
            - [callable]: a user-defined function which accepts an
                array of distances, and returns an array of the same shape
                containing the weights.

        algorithm: Algorithm used to compute the nearest neighbors:

            - 'ball_tree' will use :class:`sklearn.neighbors.BallTree`.
            - 'brute' will use a brute-force search.
            - 'auto' will attempt to decide the most appropriate algorithm.
                based on the values passed to :meth:`fit` method.

        leaf_size: Leaf size passed to BallTree or KDTree. This can affect the
            speed of the construction and query, as well as the memory
            required to store the tree. The optimal value depends on the
            nature of the problem.
        metric: The distance metric to use for the tree. The default metric is
            the L2 distance. See the documentation of the metrics module
            for a list of available metrics.
        outlier_label:
            Label, which is given for outlier samples (samples with no
            neighbors on given radius).
            If set to None, ValueError is raised, when outlier is detected.
        n_jobs: The number of parallel jobs to run for neighbors search.
            ``None`` means 1 unless in a :obj:`joblib.parallel_backend`
            context.
            ``-1`` means using all processors.

    Examples:
        Firstly, we will create a toy dataset with 2 classes.

        >>> from skfda.datasets import make_sinusoidal_process
        >>> fd1 = make_sinusoidal_process(phase_std=.25, random_state=0)
        >>> fd2 = make_sinusoidal_process(phase_mean=1.8, error_std=0.,
        ...                               phase_std=.25, random_state=0)
        >>> fd = fd1.concatenate(fd2)
        >>> y = 15*[0] + 15*[1]

        We will fit a Radius Nearest Neighbors classifier.

        >>> from skfda.ml.classification import RadiusNeighborsClassifier
        >>> neigh = RadiusNeighborsClassifier(radius=.3)
        >>> neigh.fit(fd, y)
        RadiusNeighborsClassifier(...radius=0.3...)

        We can predict the class of new samples.

        >>> neigh.predict(fd[::2]) # Predict labels for even samples
        array([0, 0, 0, 0, 0, 0, 0, 0, 1, 1, 1, 1, 1, 1, 1])

    See also:
        :class:`~skfda.ml.classification.KNeighborsClassifier`
        :class:`~skfda.ml.classification.NearestCentroid`
        :class:`~skfda.ml.regression.KNeighborsRegressor`
        :class:`~skfda.ml.regression.RadiusNeighborsRegressor`
        :class:`~skfda.ml.clustering.NearestNeighbors`

    Notes:
        See Nearest Neighbors in the sklearn online documentation for a
        discussion of the choice of ``algorithm`` and ``leaf_size``.

        This class wraps the sklearn classifier
        `sklearn.neighbors.RadiusNeighborsClassifier`.

        https://en.wikipedia.org/wiki/K-nearest_neighbor_algorithm
    """

    @overload
    def __init__(
        self: RadiusNeighborsClassifier[NDArrayFloat],
        *,
        radius: float = 1.0,
        weights: WeightsType = 'uniform',
        algorithm: AlgorithmType = 'auto',
        leaf_size: int = 30,
        metric: Literal["precomputed"],
        outlier_label: OutlierLabelType = None,
        n_jobs: int | None = None,
    ) -> None:
        pass

    @overload
    def __init__(
        self: RadiusNeighborsClassifier[InputBound],
        *,
        radius: float = 1.0,
        weights: WeightsType = 'uniform',
        algorithm: AlgorithmType = 'auto',
        leaf_size: int = 30,
        outlier_label: OutlierLabelType = None,
        n_jobs: int | None = None,
    ) -> None:
        pass

    @overload
    def __init__(
        self,
        *,
        radius: float = 1.0,
        weights: WeightsType = 'uniform',
        algorithm: AlgorithmType = 'auto',
        leaf_size: int = 30,
        metric: Metric[Input] = l2_distance,
        outlier_label: OutlierLabelType = None,
        n_jobs: int | None = None,
    ) -> None:
        pass

    def __init__(
        self,
        *,
        radius: float = 1.0,
        weights: WeightsType = 'uniform',
        algorithm: AlgorithmType = 'auto',
        leaf_size: int = 30,
        metric: Literal["precomputed"] | Metric[Input] = l2_distance,
        outlier_label: OutlierLabelType = None,
        n_jobs: int | None = None,
    ) -> None:
        super().__init__(
            radius=radius,
            weights=weights,
            algorithm=algorithm,
            leaf_size=leaf_size,
            metric=metric,
            n_jobs=n_jobs,
        )

        self.outlier_label = outlier_label

    def _init_estimator(self) -> _RadiusNeighborsClassifier:
        return _RadiusNeighborsClassifier(
            radius=self.radius,
            weights=self.weights,
            algorithm=self.algorithm,
            leaf_size=self.leaf_size,
            metric="precomputed",
            outlier_label=self.outlier_label,
            n_jobs=self.n_jobs,
        )



"""Covariance estimation."""

from typing import TYPE_CHECKING

import lazy_loader as lazy

__getattr__, __dir__, __all__ = lazy.attach(
    __name__,
    submod_attrs={
        "_base": ["CovarianceEstimator"],
        "_empirical": ["EmpiricalCovariance"],
        "_parametric_gaussian": ["ParametricGaussianCovariance"],
    },
)

if TYPE_CHECKING:
    from ._base import CovarianceEstimator as CovarianceEstimator
    from ._empirical import EmpiricalCovariance as EmpiricalCovariance
    from ._parametric_gaussian import (
        ParametricGaussianCovariance as ParametricGaussianCovariance,
    )



"""Function transformers for feature construction techniques."""
from __future__ import annotations

from typing import Optional, Sequence, Tuple

from typing_extensions import Literal

from ..._utils._sklearn_adapter import BaseEstimator, TransformerMixin
from ...representation import FData
from ...representation.grid import FDataGrid
from ...typing._base import DomainRangeLike
from ...typing._numpy import ArrayLike, NDArrayFloat, NDArrayInt
from ._functions import local_averages, number_crossings, occupation_measure


class LocalAveragesTransformer(
    BaseEstimator,
    TransformerMixin[FData, NDArrayFloat, object],
):
    r"""
    Transforms functional data to its local averages.

    It takes functional data and performs the following map:

    .. math::
        f_1(X) = \frac{1}{|T_1|} \int_{T_1} X(t) dt,\dots, \\
        f_p(X) = \frac{1}{|T_p|} \int_{T_p} X(t) dt

    where :math:`T_1, \dots, T_p` are subregions of the original
    :term:`domain`.

    Args:
        domains: Domains for each local average. It is possible to
            pass a number or a list of numbers to automatically split
            each dimension in that number of intervals and use them for
            the averages.

    See also:
        :func:`local_averages`

    Examples:
        We import the Berkeley Growth Study dataset.
        We will use only the first 3 samples to make the
        example easy.

        >>> from skfda.datasets import fetch_growth
        >>> dataset = fetch_growth(return_X_y=True)[0]
        >>> X = dataset[:3]

        We can choose the intervals used for the local averages. For example,
        we could in this case use the averages at different stages of
        development of the child: from 1 to 3 years, from 3 to 10 and from
        10 to 18:

        >>> import numpy as np
        >>> from skfda.preprocessing.feature_construction import (
        ...     LocalAveragesTransformer,
        ... )
        >>> local_averages = LocalAveragesTransformer(
        ...     domains=[(1, 3), (3, 10), (10, 18)],
        ... )
        >>> np.round(local_averages.fit_transform(X), decimals=2)
        array([[  91.37,  126.52,  179.02],
               [  87.51,  120.71,  158.81],
               [  86.36,  115.04,  156.37]])

        A different possibility is to decide how many intervals we want to
        consider.  For example, we could want to split the domain in 2
        intervals of the same length.

        >>> local_averages = LocalAveragesTransformer(domains=2)
        >>> np.around(local_averages.fit_transform(X), decimals=2)
        array([[ 116.94,  177.26],
               [ 111.86,  157.62],
               [ 107.29,  154.97]])
    """

    def __init__(
        self,
        *,
        domains: int | Sequence[int] | Sequence[DomainRangeLike],
    ) -> None:
        self.domains = domains

    def transform(self, X: FData, y: object = None) -> NDArrayFloat:
        """
        Transform the provided data to its local averages.

        Args:
            X: FDataGrid with the samples that are going to be transformed.
            y: Unused.

        Returns:
            Array of shape (n_samples, n_intervals) including
            the transformed data.

        """
        return local_averages(
            X,
            domains=self.domains,
        ).reshape(X.data_matrix.shape[0], -1)


class OccupationMeasureTransformer(
    BaseEstimator,
    TransformerMixin[FData, NDArrayFloat, object],
):
    """
    Transformer that works as an adapter for the occupation_measure function.

    Args:
        intervals: ndarray of tuples containing the
            intervals we want to consider. The shape should be
            (n_sequences, 2)
        n_points: Number of points to evaluate in the domain.
            By default will be used the points defined on the FDataGrid.
            On a FDataBasis this value should be specified.

    Example:
        We will create the FDataGrid that we will use to extract
        the occupation measure
        >>> from skfda.representation import FDataGrid
        >>> import numpy as np
        >>> t = np.linspace(0, 10, 100)
        >>> fd_grid = FDataGrid(
        ...     data_matrix=[
        ...         t,
        ...         2 * t,
        ...         np.sin(t),
        ...     ],
        ...     grid_points=t,
        ... )

        Finally we call to the occupation measure function with the
        intervals that we want to consider. In our case (0.0, 1.0)
        and (2.0, 3.0). We need also to specify the number of points
        we want that the function takes into account to interpolate.
        We are going to use 501 points.
        >>> from skfda.preprocessing.feature_construction import (
        ...     OccupationMeasureTransformer,
        ... )
        >>> occupation_measure = OccupationMeasureTransformer(
        ...     intervals=[(0.0, 1.0), (2.0, 3.0)],
        ...     n_points=501,
        ... )

        >>> np.around(occupation_measure.fit_transform(fd_grid), decimals=2)
        array([[ 0.98,  1.  ],
               [ 0.5 ,  0.52],
               [ 6.28,  0.  ]])
    """

    def __init__(
        self,
        intervals: Sequence[Tuple[float, float]],
        *,
        n_points: Optional[int] = None,
    ):
        self.intervals = intervals
        self.n_points = n_points

    def transform(self, X: FData, y: object = None) -> NDArrayFloat:
        """
        Transform the provided data using the occupation_measure function.

        Args:
            X: FDataGrid or FDataBasis with the samples that are going to be
                transformed.

        Returns:
            Array of shape (n_intervals, n_samples) including the transformed
            data.
        """
        return occupation_measure(X, self.intervals, n_points=self.n_points)


class NumberCrossingsTransformer(
    BaseEstimator,
    TransformerMixin[FDataGrid, NDArrayInt, object],
):
    """
    Transformer that works as an adapter for the number_up_crossings function.

    Args:
        levels: Sequence of numbers including the levels
            we want to consider for the crossings. By
            default it calculates zero-crossings.
        direction: Whether to consider only up-crossings,
            down-crossings or both.

    Example:
        For this example we will use a well known function so the correct
        functioning of this method can be checked.
        We will create and use a DataFrame with a sample extracted from
        the Bessel Function of first type and order 0.
        First of all we import the Bessel Function and create the X axis
        data grid. Then we create the FdataGrid.
        >>> from skfda.preprocessing.feature_construction import (
        ...     NumberCrossingsTransformer,
        ... )
        >>> from scipy.special import jv
        >>> from skfda.representation import FDataGrid
        >>> import numpy as np
        >>> x_grid = np.linspace(0, 14, 14)
        >>> fd_grid = FDataGrid(
        ...     data_matrix=[jv([0], x_grid)],
        ...     grid_points=x_grid,
        ... )
        >>> fd_grid.data_matrix
        array([[[ 1.        ],
        [ 0.73041066],
        [ 0.13616752],
        [-0.32803875],
        [-0.35967936],
        [-0.04652559],
        [ 0.25396879],
        [ 0.26095573],
        [ 0.01042895],
        [-0.22089135],
        [-0.2074856 ],
        [ 0.0126612 ],
        [ 0.20089319],
        [ 0.17107348]]])

        Finally we evaluate the number of zero-upcrossings method with the
        FDataGrid created.
        >>> tf = NumberCrossingsTransformer(levels=0, direction="up")
        >>> tf.fit_transform(fd_grid)
        array([[2]])
    """

    def __init__(
        self,
        *,
        levels: ArrayLike = 0,
        direction: Literal["up", "down", "all"] = "all",
    ):
        self.levels = levels
        self.direction = direction

    def transform(self, X: FDataGrid, y: object = None) -> NDArrayInt:
        """
        Transform the provided data using the number_up_crossings function.

        Args:
            X: FDataGrid with the samples that are going to be transformed.

        Returns:
            Array of shape (n_samples, len(levels)) including the transformed
            data.
        """
        return number_crossings(
            X,
            levels=self.levels,
            direction=self.direction,
        )



"""Module for functional data manipulation.

Defines the abstract class that should be implemented by the funtional data
objects of the package and contains some commons methods.
"""

from __future__ import annotations

import warnings
from abc import ABC, abstractmethod
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Iterable,
    Iterator,
    NoReturn,
    Optional,
    Sequence,
    TypeVar,
    Union,
    cast,
    overload,
)

import numpy as np
import pandas.api.extensions
from matplotlib.figure import Figure
from typing_extensions import Literal, Protocol

from .._utils import _evaluate_grid, _to_grid_points
from ..typing._base import (
    DomainRange,
    GridPointsLike,
    LabelTuple,
    LabelTupleLike,
)
from ..typing._numpy import (
    ArrayLike,
    NDArrayBool,
    NDArrayFloat,
    NDArrayInt,
    NDArrayObject,
)
from .evaluator import Evaluator
from .extrapolation import ExtrapolationLike, _parse_extrapolation

if TYPE_CHECKING:
    from .basis import Basis, FDataBasis
    from .grid import FDataGrid

T = TypeVar('T', bound='FData')

EvalPointsType = Union[
    ArrayLike,
    GridPointsLike,
    Iterable[GridPointsLike],
]


class FData(  # noqa: WPS214
    ABC,
    pandas.api.extensions.ExtensionArray,  # type: ignore[misc]
):
    """Defines the structure of a functional data object.

    Attributes:
        n_samples (int): Number of samples.
        dim_domain (int): Dimension of the domain.
        dim_codomain (int): Dimension of the image.
        extrapolation (Extrapolation): Default extrapolation mode.
        dataset_name (str): name of the dataset.
        argument_names (tuple): tuple containing the names of the different
            arguments.
        coordinate_names (tuple): tuple containing the names of the different
            coordinate functions.

    """

    dataset_name: Optional[str]

    def __init__(
        self,
        *,
        extrapolation: Optional[ExtrapolationLike] = None,
        dataset_name: Optional[str] = None,
        argument_names: Optional[LabelTupleLike] = None,
        coordinate_names: Optional[LabelTupleLike] = None,
        sample_names: Optional[LabelTupleLike] = None,
    ) -> None:

        self.extrapolation = extrapolation  # type: ignore[assignment]
        self.dataset_name = dataset_name

        self.argument_names = argument_names  # type: ignore[assignment]
        self.coordinate_names = coordinate_names  # type: ignore[assignment]
        self.sample_names = sample_names  # type: ignore[assignment]

    @property
    def argument_names(self) -> LabelTuple:
        return self._argument_names

    @argument_names.setter
    def argument_names(
        self,
        names: Optional[LabelTupleLike],
    ) -> None:
        if names is None:
            names = (None,) * self.dim_domain
        else:
            names = tuple(names)
            if len(names) != self.dim_domain:
                raise ValueError(
                    "There must be a name for each of the "
                    "dimensions of the domain.",
                )

        self._argument_names = names

    @property
    def coordinate_names(self) -> LabelTuple:
        return self._coordinate_names

    @coordinate_names.setter
    def coordinate_names(
        self,
        names: Optional[LabelTupleLike],
    ) -> None:
        if names is None:
            names = (None,) * self.dim_codomain
        else:
            names = tuple(names)
            if len(names) != self.dim_codomain:
                raise ValueError(
                    "There must be a name for each of the "
                    "dimensions of the codomain.",
                )

        self._coordinate_names = names

    @property
    def sample_names(self) -> LabelTuple:
        return self._sample_names

    @sample_names.setter
    def sample_names(self, names: Optional[LabelTupleLike]) -> None:
        if names is None:
            names = (None,) * self.n_samples
        else:
            names = tuple(names)
            if len(names) != self.n_samples:
                raise ValueError(
                    "There must be a name for each of the samples.",
                )

        self._sample_names = names

    @property
    @abstractmethod
    def n_samples(self) -> int:
        """Return the number of samples.

        Returns:
            Number of samples of the FData object.

        """
        pass

    @property
    @abstractmethod
    def dim_domain(self) -> int:
        """Return number of dimensions of the :term:`domain`.

        Returns:
            Number of dimensions of the domain.

        """
        pass

    @property
    @abstractmethod
    def dim_codomain(self) -> int:
        """Return number of dimensions of the :term:`codomain`.

        Returns:
            Number of dimensions of the codomain.

        """
        pass

    @property
    @abstractmethod
    def coordinates(self: T) -> _CoordinateSequence:
        r"""Return a component of the FDataGrid.

        If the functional object contains multivariate samples
        :math:`f: \mathbb{R}^n \rightarrow \mathbb{R}^d`, this method returns
        an iterator of the vector :math:`f = (f_1, ..., f_d)`.

        """
        pass

    @property
    def extrapolation(self) -> Optional[Evaluator]:
        """Return default type of extrapolation."""
        return self._extrapolation

    @extrapolation.setter
    def extrapolation(self, value: Optional[ExtrapolationLike]) -> None:
        """Set the type of extrapolation."""
        self._extrapolation = _parse_extrapolation(value)

    @property
    @abstractmethod
    def domain_range(self) -> DomainRange:
        """Return the :term:`domain` range of the object

        Returns:
            List of tuples with the ranges for each domain dimension.
        """
        pass

    def _extrapolation_index(self, eval_points: NDArrayFloat) -> NDArrayBool:
        """Check the points that need to be extrapolated.

        Args:
            eval_points: Array with shape `n_eval_points` x
                `dim_domain` with the evaluation points, or shape ´n_samples´ x
                `n_eval_points` x `dim_domain` with different evaluation
                points for each sample.

        Returns:
            Array with boolean index. The positions with True
            in the index are outside the domain range and extrapolation
            should be applied.

        """
        index = np.zeros(eval_points.shape[:-1], dtype=np.bool_)

        # Checks bounds in each domain dimension
        for i, bounds in enumerate(self.domain_range):
            np.logical_or(index, eval_points[..., i] < bounds[0], out=index)
            np.logical_or(index, eval_points[..., i] > bounds[1], out=index)

        return index

    def _join_evaluation(
        self,
        index_matrix: NDArrayBool,
        index_ext: NDArrayBool,
        index_ev: NDArrayBool,
        res_extrapolation: NDArrayFloat,
        res_evaluation: NDArrayFloat,
    ) -> NDArrayFloat:
        """Join the points evaluated.

        This method is used internally by :func:`evaluate` to join the result
        of the evaluation and the result of the extrapolation.

        Args:
            index_matrix: Boolean index with the points extrapolated.
            index_ext: Boolean index with the columns that contains
                points extrapolated.
            index_ev: Boolean index with the columns that contains
                points evaluated.
            res_extrapolation: Result of the extrapolation.
            res_evaluation: Result of the evaluation.

        Returns:
            Matrix with the points evaluated with shape
            `n_samples` x `number of points evaluated` x `dim_codomain`.

        """
        res = np.empty((
            self.n_samples,
            index_matrix.shape[-1],
            self.dim_codomain,
        ))

        # Case aligned evaluation
        if index_matrix.ndim == 1:
            res[:, index_ev, :] = res_evaluation
            res[:, index_ext, :] = res_extrapolation

        else:

            res[:, index_ev] = res_evaluation
            res[index_matrix] = res_extrapolation[index_matrix[:, index_ext]]

        return res

    @abstractmethod
    def _evaluate(
        self,
        eval_points: NDArrayFloat,
        *,
        aligned: bool = True,
    ) -> NDArrayFloat:
        """Define the evaluation of the FData.

        Evaluates the samples of an FData object at several points.

        Subclasses must override this method to implement evaluation.

        Args:
            eval_points: List of points where the functions are
                evaluated. If `aligned` is `True`, then a list of
                lists of points must be passed, with one list per sample.
            aligned: Whether the input points are
                the same for each sample, or an array of points per sample is
                passed.

        Returns:
            Numpy 3d array with shape `(n_samples,
            len(eval_points), dim_codomain)` with the result of the
            evaluation. The entry (i,j,k) will contain the value k-th image
            dimension of the i-th sample, at the j-th evaluation point.

        """
        pass

    @overload
    def evaluate(
        self,
        eval_points: ArrayLike,
        *,
        derivative: int = 0,
        extrapolation: Optional[ExtrapolationLike] = None,
        grid: Literal[False] = False,
        aligned: Literal[True] = True,
    ) -> NDArrayFloat:
        pass

    @overload
    def evaluate(
        self,
        eval_points: Iterable[ArrayLike],
        *,
        derivative: int = 0,
        extrapolation: Optional[ExtrapolationLike] = None,
        grid: Literal[False] = False,
        aligned: Literal[False],
    ) -> NDArrayFloat:
        pass

    @overload
    def evaluate(
        self,
        eval_points: GridPointsLike,
        *,
        derivative: int = 0,
        extrapolation: Optional[ExtrapolationLike] = None,
        grid: Literal[True],
        aligned: Literal[True] = True,
    ) -> NDArrayFloat:
        pass

    @overload
    def evaluate(
        self,
        eval_points: Iterable[GridPointsLike],
        *,
        derivative: int = 0,
        extrapolation: Optional[ExtrapolationLike] = None,
        grid: Literal[True],
        aligned: Literal[False],
    ) -> NDArrayFloat:
        pass

    def evaluate(
        self,
        eval_points: EvalPointsType,
        *,
        derivative: int = 0,
        extrapolation: Optional[ExtrapolationLike] = None,
        grid: bool = False,
        aligned: bool = True,
    ) -> NDArrayFloat:
        """
        Evaluate the object at a list of values or a grid.

        ..  deprecated:: 0.8
            Use normal calling notation instead.

        Args:
            eval_points: List of points where the functions are
                evaluated. If ``grid`` is ``True``, a list of axes, one per
                :term:`domain` dimension, must be passed instead. If
                ``aligned`` is ``True``, then a list of lists (of points or
                axes, as explained) must be passed, with one list per sample.
            derivative: Deprecated. Order of the derivative to evaluate.
            extrapolation: Controls the
                extrapolation mode for elements outside the domain range. By
                default it is used the mode defined during the instance of the
                object.
            grid: Whether to evaluate the results on a grid
                spanned by the input arrays, or at points specified by the
                input arrays. If true the eval_points should be a list of size
                dim_domain with the corresponding times for each axis. The
                return matrix has shape n_samples x len(t1) x len(t2) x ... x
                len(t_dim_domain) x dim_codomain. If the domain dimension is 1
                the parameter has no efect. Defaults to False.
            aligned: Whether the input points are the same for each sample,
                or an array of points per sample is passed.

        Returns:
            Matrix whose rows are the values of the each
            function at the values specified in eval_points.

        """
        warnings.warn(
            "The method 'evaluate' is deprecated. "
            "Please use the normal calling notation on the functional data "
            "object instead.",
            DeprecationWarning,
            stacklevel=2,
        )

        return self(
            eval_points=eval_points,
            derivative=derivative,
            extrapolation=extrapolation,
            grid=grid,
            aligned=aligned,
        )

    @overload
    def __call__(
        self,
        eval_points: ArrayLike,
        *,
        derivative: int = 0,
        extrapolation: Optional[ExtrapolationLike] = None,
        grid: Literal[False] = False,
        aligned: bool = True,
    ) -> NDArrayFloat:
        pass

    @overload
    def __call__(
        self,
        eval_points: GridPointsLike,
        *,
        derivative: int = 0,
        extrapolation: Optional[ExtrapolationLike] = None,
        grid: Literal[True],
        aligned: Literal[True] = True,
    ) -> NDArrayFloat:
        pass

    @overload
    def __call__(
        self,
        eval_points: Iterable[GridPointsLike],
        *,
        derivative: int = 0,
        extrapolation: Optional[ExtrapolationLike] = None,
        grid: Literal[True],
        aligned: Literal[False],
    ) -> NDArrayFloat:
        pass

    @overload
    def __call__(
        self,
        eval_points: EvalPointsType,
        *,
        derivative: int = 0,
        extrapolation: Optional[ExtrapolationLike] = None,
        grid: bool = False,
        aligned: bool = True,
    ) -> NDArrayFloat:
        pass

    def __call__(
        self,
        eval_points: EvalPointsType,
        *,
        derivative: int = 0,
        extrapolation: Optional[ExtrapolationLike] = None,
        grid: bool = False,
        aligned: bool = True,
    ) -> NDArrayFloat:
        """
        Evaluate the :term:`functional object`.

        Evaluate the object or its derivatives at a list of values or a
        grid. This method is a wrapper of :meth:`evaluate`.

        Args:
            eval_points: List of points where the functions are
                evaluated. If a matrix of shape nsample x eval_points is given
                each sample is evaluated at the values in the corresponding row
                in eval_points.
            derivative: Order of the derivative. Defaults to 0.
            extrapolation: Controls the
                extrapolation mode for elements outside the domain range. By
                default it is used the mode defined during the instance of the
                object.
            grid: Whether to evaluate the results on a grid
                spanned by the input arrays, or at points specified by the
                input arrays. If true the eval_points should be a list of size
                dim_domain with the corresponding times for each axis. The
                return matrix has shape n_samples x len(t1) x len(t2) x ... x
                len(t_dim_domain) x dim_codomain. If the domain dimension is 1
                the parameter has no efect. Defaults to False.
            aligned: Whether the input points are the same for each sample,
                or an array of points per sample is passed.

        Returns:
            Matrix whose rows are the values of the each
            function at the values specified in eval_points.

        """
        from ..misc.validation import validate_evaluation_points

        if derivative != 0:
            warnings.warn(
                "Parameter derivative is deprecated. Use the "
                "derivative function instead.",
                DeprecationWarning,
            )
            return self.derivative(order=derivative)(
                eval_points,
                extrapolation=extrapolation,
                grid=grid,
                aligned=aligned,
            )

        if grid:  # Evaluation of a grid performed in auxiliar function

            return _evaluate_grid(
                eval_points,
                evaluate_method=self,
                n_samples=self.n_samples,
                dim_domain=self.dim_domain,
                dim_codomain=self.dim_codomain,
                extrapolation=extrapolation,
                aligned=aligned,
            )

        if extrapolation is None:
            extrapolation = self.extrapolation
        else:
            # Gets the function to perform extrapolation or None
            extrapolation = _parse_extrapolation(extrapolation)

        eval_points = cast(
            ArrayLike,
            eval_points,
        )

        # Convert to array and check dimensions of eval points
        eval_points = validate_evaluation_points(
            eval_points,
            aligned=aligned,
            n_samples=self.n_samples,
            dim_domain=self.dim_domain,
        )

        if extrapolation is not None:

            index_matrix = self._extrapolation_index(eval_points)

            if index_matrix.any():

                # Partition of eval points
                if aligned:

                    index_ext = index_matrix
                    index_ev = ~index_matrix

                    eval_points_extrapolation = eval_points[index_ext]
                    eval_points_evaluation = eval_points[index_ev]

                else:
                    index_ext = np.logical_or.reduce(index_matrix, axis=0)
                    eval_points_extrapolation = eval_points[:, index_ext]

                    index_ev = np.logical_or.reduce(~index_matrix, axis=0)
                    eval_points_evaluation = eval_points[:, index_ev]

                # Direct evaluation
                res_evaluation = self._evaluate(
                    eval_points_evaluation,
                    aligned=aligned,
                )

                res_extrapolation = extrapolation(
                    self,
                    eval_points_extrapolation,
                    aligned=aligned,
                )

                return self._join_evaluation(
                    index_matrix,
                    index_ext,
                    index_ev,
                    res_extrapolation,
                    res_evaluation,
                )

        return self._evaluate(
            eval_points,
            aligned=aligned,
        )

    @abstractmethod
    def derivative(self: T, *, order: int = 1) -> T:
        """Differentiate a FData object.

        Args:
            order: Order of the derivative. Defaults to one.

        Returns:
            Functional object containg the derivative.

        """
        pass

    @abstractmethod
    def integrate(
        self: T,
        *,
        domain: Optional[DomainRange] = None,
    ) -> NDArrayFloat:
        """
        Integration of the FData object.

        The integration is performed over the whole domain. Thus, for a
        function of several variables this will be a multiple integral.

        For a vector valued function the vector of integrals will be
        returned.

        Args:
            domain: Domain range where we want to integrate.
                By default is None as we integrate on the whole domain.

        Returns:
            NumPy array of size (``n_samples``, ``dim_codomain``)
            with the integrated data.

        """
        pass

    @abstractmethod
    def shift(
        self,
        shifts: Union[ArrayLike, float],
        *,
        restrict_domain: bool = False,
        extrapolation: Optional[ExtrapolationLike] = None,
        grid_points: Optional[GridPointsLike] = None,
    ) -> FDataGrid:
        r"""
        Perform a shift of the curves.

        The i-th shifted function :math:`y_i` has the form

        .. math::
            y_i(t) = x_i(t + \delta_i)

        where :math:`x_i` is the i-th original function and :math:`delta_i` is
        the shift performed for that function, that must be a vector in the
        domain space.

        Note that a positive shift moves the graph of the function in the
        negative direction and vice versa.

        Args:
            shifts: List with the shifts
                corresponding for each sample or numeric with the shift to
                apply to all samples.
            restrict_domain: If True restricts the domain to avoid the
                evaluation of points outside the domain using extrapolation.
                Defaults uses extrapolation.
            extrapolation: Controls the
                extrapolation mode for elements outside the domain range.
                By default uses the method defined in fd. See extrapolation to
                more information.
            grid_points: Grid of points where
                the functions are evaluated to obtain the discrete
                representation of the object to operate. If ``None`` the
                current grid_points are used to unificate the domain of the
                shifted data.

        Returns:
            Shifted functions.

        """
        assert grid_points is not None
        grid_points = _to_grid_points(grid_points)

        arr_shifts = np.array([shifts] if np.isscalar(shifts) else shifts)

        # Accept unidimensional array when the domain dimension is one or when
        # the shift is the same for each sample
        if arr_shifts.ndim == 1:
            arr_shifts = (
                arr_shifts[np.newaxis, :]  # Same shift for each sample
                if len(arr_shifts) == self.dim_domain
                else arr_shifts[:, np.newaxis]
            )

        if len(arr_shifts) not in {1, self.n_samples}:
            raise ValueError(
                f"The length of the shift vector ({len(arr_shifts)}) must "
                f"have length equal to 1 or to the number of samples "
                f"({self.n_samples})",
            )

        if restrict_domain:
            domain = np.asarray(self.domain_range)

            a = domain[:, 0] - np.min(np.min(arr_shifts, axis=0), 0)
            b = domain[:, 1] - np.max(np.max(arr_shifts, axis=1), 0)

            domain = np.hstack((a, b))
            domain_range = tuple(domain)

        if len(arr_shifts) == 1:
            shifted_grid_points = tuple(
                g + s for g, s in zip(grid_points, arr_shifts[0])
            )
            data_matrix = self(
                shifted_grid_points,
                extrapolation=extrapolation,
                aligned=True,
                grid=True,
            )
        else:
            shifted_grid_points_per_sample = (
                tuple(
                    g + s for g, s in zip(grid_points, shift)
                ) for shift in arr_shifts
            )
            data_matrix = self(
                shifted_grid_points_per_sample,
                extrapolation=extrapolation,
                aligned=False,
                grid=True,
            )

        shifted = self.to_grid().copy(
            data_matrix=data_matrix,
            grid_points=grid_points,
        )

        if restrict_domain:
            shifted = shifted.restrict(domain_range)

        return shifted

    def plot(self, *args: Any, **kwargs: Any) -> Figure:
        """Plot the FDatGrid object.

        Args:
            args: Positional arguments to be passed to the class
                :class:`~skfda.exploratory.visualization.representation.GraphPlot`.
            kwargs: Keyword arguments to be passed to the class
                :class:`~skfda.exploratory.visualization.representation.GraphPlot`.

        Returns:
            Figure object in which the graphs are plotted.

        """
        from ..exploratory.visualization.representation import GraphPlot

        return GraphPlot(self, *args, **kwargs).plot()

    @abstractmethod
    def copy(
        self: T,
        *,
        deep: bool = False,  # For Pandas compatibility
        dataset_name: Optional[str] = None,
        argument_names: Optional[LabelTupleLike] = None,
        coordinate_names: Optional[LabelTupleLike] = None,
        sample_names: Optional[LabelTupleLike] = None,
        extrapolation: Optional[ExtrapolationLike] = None,
    ) -> T:
        """Make a copy of the object."""
        pass

    @abstractmethod  # noqa: WPS125
    def sum(  # noqa: WPS125
        self: T,
        *,
        axis: int | None = None,
        out: None = None,
        keepdims: bool = False,
        skipna: bool = False,
        min_count: int = 0,
    ) -> T:
        """Compute the sum of all the samples.

        Args:
            axis: Used for compatibility with numpy. Must be None or 0.
            out: Used for compatibility with numpy. Must be None.
            keepdims: Used for compatibility with numpy. Must be False.
            skipna: Wether the NaNs are ignored or not.
            min_count: Number of valid (non NaN) data to have in order
                for the a variable to not be NaN when `skipna` is
                `True`.

        Returns:
            A FData object with just one sample representing
            the sum of all the samples in the original object.

        """
        if (
            (axis is not None and axis != 0)
            or out is not None
            or keepdims is not False
        ):
            raise NotImplementedError(
                "Not implemented for that parameter combination",
            )

        return self

    @overload
    def cov(  # noqa: WPS451
        self: T,
        s_points: NDArrayFloat,
        t_points: NDArrayFloat,
        /,
        correction: int = 0,
    ) -> NDArrayFloat:
        pass

    @overload
    def cov(  # noqa: WPS451
        self: T,
        /,
        correction: int = 0,
    ) -> Callable[[NDArrayFloat, NDArrayFloat], NDArrayFloat]:
        pass

    @abstractmethod
    def cov(  # noqa: WPS320, WPS451
        self: T,
        s_points: Optional[NDArrayFloat] = None,
        t_points: Optional[NDArrayFloat] = None,
        /,
        correction: int = 0,
    ) -> Union[
        Callable[[NDArrayFloat, NDArrayFloat], NDArrayFloat],
        NDArrayFloat,
    ]:
        """Compute the covariance of the functional data object.

        Calculates the unbiased sample covariance function of the data.
        This is expected to be only defined for univariate functions.
        The resulting covariance function is defined in the cartesian
        product of the domain of the functions.
        If s_points or t_points are not provided, this method returns
        a callable object representing the covariance function.
        If s_points and t_points are provided, this method returns the
        evaluation of the covariance function at the grid formed by the
        cartesian product of the points in s_points and t_points.

        Args:
            s_points: Points where the covariance function is evaluated.
            t_points: Points where the covariance function is evaluated.
            correction: degrees of freedom adjustment. The divisor used in the
                calculation is `N - correction`, where `N` represents the
                number of elements. Default: `0`.

        Returns:
            Covariance function.

        """
        pass

    def mean(
        self: T,
        *,
        axis: int | None = None,
        dtype: None = None,
        out: None = None,
        keepdims: bool = False,
        skipna: bool = False,
    ) -> T:
        """Compute the mean of all the samples.

        Args:
            axis: Used for compatibility with numpy. Must be None or 0.
            dtype: Used for compatibility with numpy. Must be None.
            out: Used for compatibility with numpy. Must be None.
            keepdims: Used for compatibility with numpy. Must be False.
            skipna: Wether the NaNs are ignored or not.

        Returns:
            A FData object with just one sample representing
            the mean of all the samples in the original object.

        """
        if dtype is not None:
            raise NotImplementedError(
                "Not implemented for that parameter combination",
            )

        return (
            self.sum(axis=axis, out=out, keepdims=keepdims, skipna=skipna)
            / self.n_samples
        )

    @abstractmethod
    def to_grid(
        self,
        grid_points: Optional[GridPointsLike] = None,
    ) -> FDataGrid:
        """Return the discrete representation of the object.

        Args:
            grid_points: Points per axis
                where the function is going to be evaluated.

        Returns:
            Discrete representation of the functional data
            object.

        """
        pass

    @abstractmethod
    def to_basis(
        self,
        basis: Basis,
        **kwargs: Any,
    ) -> FDataBasis:
        """Return the basis representation of the object.

        Args:
            basis: basis object in which the functional data are
                going to be represented.
            kwargs: keyword arguments to be passed to
                FDataBasis.from_data().

        Returns:
            Basis representation of the funtional data
            object.

        """
        pass

    @abstractmethod
    def concatenate(self: T, *others: T, as_coordinates: bool = False) -> T:
        """Join samples from a similar FData object.

        Joins samples from another FData object if it has the same
        dimensions and has compatible representations.

        Args:
            others: other FData objects.
            as_coordinates:  If False concatenates as
                new samples, else, concatenates the other functions as
                new components of the image. Defaults to False.

        Returns:
            :class:`FData`: FData object with the samples from the two
            original objects.

        """
        pass

    @abstractmethod
    def compose(
        self: T,
        fd: T,
        *,
        eval_points: Optional[NDArrayFloat] = None,
    ) -> FData:
        """Composition of functions.

        Performs the composition of functions.

        Args:
            fd: FData object to make the composition. Should
                have the same number of samples and image dimension equal to
                the domain dimension of the object composed.
            eval_points: Points to perform the evaluation.

        """
        pass

    @abstractmethod
    def __getitem__(
        self: T,
        key: Union[int, slice, NDArrayInt],
    ) -> T:
        """Return self[key]."""
        pass

    def equals(self, other: object) -> bool:
        """Whole object equality."""
        return (
            isinstance(other, type(self))  # noqa: WPS222
            and self.extrapolation == other.extrapolation
            and self.dataset_name == other.dataset_name
            and self.argument_names == other.argument_names
            and self.coordinate_names == other.coordinate_names
        )

    @abstractmethod
    def _eq_elemenwise(self: T, other: T) -> NDArrayBool:
        """Elementwise equality."""
        pass

    def __eq__(self, other: object) -> NDArrayBool:  # type: ignore[override]
        """Elementwise equality, as with arrays."""
        if not isinstance(other, type(self)) or self.dtype != other.dtype:
            if other is pandas.NA:
                return self.isna()
            if pandas.api.types.is_list_like(other) and not isinstance(
                other, (pandas.Series, pandas.Index, pandas.DataFrame),
            ):
                other = cast(Iterable[object], other)
                return np.concatenate([x == y for x, y in zip(self, other)])

            return NotImplemented

        if len(self) != len(other) and len(self) != 1 and len(other) != 1:
            raise ValueError(
                f"Different lengths: "
                f"len(self)={len(self)} and "
                f"len(other)={len(other)}",
            )

        return self._eq_elemenwise(other)

    def __ne__(self, other: object) -> NDArrayBool:  # type: ignore[override]
        """Return for `self != other` (element-wise in-equality)."""
        result = self.__eq__(other)
        if result is NotImplemented:
            return NotImplemented

        return ~result

    def _copy_op(
        self: T,
        other: Union[T, NDArrayFloat, NDArrayInt, float],
        **kwargs: Any,
    ) -> T:

        base_copy = (
            other if isinstance(other, type(self))
            and self.n_samples == 1 and other.n_samples != 1
            else self
        )

        return base_copy.copy(**kwargs)

    @abstractmethod
    def __add__(self: T, other: T) -> T:
        """Addition for FData object."""
        pass

    @abstractmethod
    def __radd__(self: T, other: T) -> T:
        """Addition for FData object."""
        pass

    @abstractmethod
    def __sub__(self: T, other: T) -> T:
        """Subtraction for FData object."""
        pass

    @abstractmethod
    def __rsub__(self: T, other: T) -> T:
        """Right subtraction for FData object."""
        pass

    @abstractmethod
    def __mul__(
        self: T,
        other: Union[NDArrayFloat, NDArrayInt, float],
    ) -> T:
        """Multiplication for FData object."""
        pass

    @abstractmethod
    def __rmul__(
        self: T,
        other: Union[NDArrayFloat, NDArrayInt, float],
    ) -> T:
        """Multiplication for FData object."""
        pass

    @abstractmethod
    def __truediv__(
        self: T,
        other: Union[NDArrayFloat, NDArrayInt, float],
    ) -> T:
        """Division for FData object."""
        pass

    @abstractmethod
    def __rtruediv__(
        self: T,
        other: Union[NDArrayFloat, NDArrayInt, float],
    ) -> T:
        """Right division for FData object."""
        pass

    @abstractmethod
    def __neg__(self: T) -> T:
        """Negation of FData object."""
        pass

    def __iter__(self: T) -> Iterator[T]:
        """Iterate over the samples."""
        yield from (self[i] for i in range(self.n_samples))

    def __len__(self) -> int:
        """Return the number of samples of the FData object."""
        return self.n_samples

    #####################################################################
    # Numpy methods
    #####################################################################

    def __array__(self, *args: Any, **kwargs: Any) -> NDArrayObject:
        """Return a numpy array with the objects."""
        # This is to prevent numpy to access inner dimensions
        array = np.empty(shape=len(self), dtype=np.object_)

        for i, f in enumerate(self):
            array[i] = f

        return array

    def __array_ufunc__(
        self,
        ufunc: Any,
        method: str,
        *inputs: Any,
        **kwargs: Any,
    ) -> Any:
        """Prevent NumPy from converting to array just to do operations."""
        # Make normal multiplication by scalar use the __mul__ method
        if ufunc == np.multiply and method == "__call__" and len(inputs) == 2:
            if isinstance(inputs[0], np.ndarray):
                inputs = inputs[::-1]

            return inputs[0] * inputs[1]

        return NotImplemented

    #####################################################################
    # Pandas ExtensionArray methods
    #####################################################################
    @property
    def ndim(self) -> int:
        """
        Return number of dimensions of the functional data.

        It is always 1, as each observation is considered a "scalar" object.

        Returns:
            Number of dimensions of the functional data.

        """
        return 1

    @classmethod
    def _from_sequence(
        cls,
        scalars: Union['FData', Sequence['FData']],
        dtype: Any = None,
        copy: bool = False,
    ) -> 'FData':

        scalars_seq: Sequence['FData'] = (
            [scalars] if isinstance(scalars, cls) else scalars
        )

        if copy:
            scalars_seq = [f.copy() for f in scalars_seq]

        if dtype is None:
            first_element = next(s for s in scalars_seq if s is not pandas.NA)
            dtype = first_element.dtype

        scalars_seq = [
            s if s is not pandas.NA else dtype._na_repr()  # noqa: WPS437
            for s in scalars_seq
        ]

        if len(scalars_seq) == 0:
            scalars_seq = [dtype._na_repr()[:0]]  # noqa: WPS437

        return cls._concat_same_type(scalars_seq)

    @classmethod
    def _from_factorized(cls, values: Any, original: Any) -> NoReturn:
        raise NotImplementedError(
            "Factorization does not make sense for functional data",
        )

    @abstractmethod
    def _take_allow_fill(
        self: T,
        indices: NDArrayInt,
        fill_value: T,
    ) -> T:
        pass

    @abstractmethod
    def isna(self) -> NDArrayBool:  # noqa: D102
        pass

    def take(  # noqa: WPS238
        self: T,
        indices: Union[int, Sequence[int], NDArrayInt],
        allow_fill: bool = False,
        fill_value: Optional[T] = None,
        axis: int = 0,
    ) -> T:
        """
        Take elements from an array.

        Parameters:
            indices: Indices to be taken.
            allow_fill: How to handle negative values in `indices`.

                * False: negative values in `indices` indicate positional
                  indices from the right (the default). This is similar to
                  :func:`numpy.take`.
                * True: negative values in `indices` indicate
                  missing values. These values are set to `fill_value`. Any
                  other negative values raise a ``ValueError``.

            fill_value: Fill value to use for NA-indices
                when `allow_fill` is True.
                This may be ``None``, in which case the default NA value for
                the type, ``self.dtype.na_value``, is used.
                For many ExtensionArrays, there will be two representations of
                `fill_value`: a user-facing "boxed" scalar, and a low-level
                physical NA value. `fill_value` should be the user-facing
                version, and the implementation should handle translating that
                to the physical version for processing the take if necessary.
            axis: Parameter for compatibility with numpy. Must be always 0.

        Returns:
            FData

        Raises:
            IndexError: When the indices are out of bounds for the array.
            ValueError: When `indices` contains negative values other than
                ``-1`` and `allow_fill` is True.

        Notes:
            ExtensionArray.take is called by ``Series.__getitem__``, ``.loc``,
            ``iloc``, when `indices` is a sequence of values. Additionally,
            it's called by :meth:`Series.reindex`, or any other method
            that causes realignment, with a `fill_value`.

        See Also:
            numpy.take
            pandas.api.extensions.take
        """
        # The axis parameter must exist, because sklearn tries to use take
        # instead of __getitem__
        if axis != 0:
            raise ValueError(f"Axis must be 0, not {axis}")

        arr_indices = np.atleast_1d(indices)

        if fill_value is None:
            fill_value = self.dtype.na_value

        non_empty_take_msg = "cannot do a non-empty take from an empty axes"

        if allow_fill:
            if (arr_indices < -1).any():
                raise ValueError("Invalid indexes")

            positive_mask = arr_indices >= 0
            if len(self) == 0 and positive_mask.any():
                raise IndexError(non_empty_take_msg)

            sample_names = np.zeros(len(arr_indices), dtype=object)
            result = self._take_allow_fill(arr_indices, fill_value)

            sample_names[positive_mask] = np.array(self.sample_names)[
                arr_indices[positive_mask]
            ]

            if fill_value is not self.dtype.na_value:
                sample_names[~positive_mask] = fill_value.sample_names[0]

            result.sample_names = tuple(sample_names)
        else:
            if len(self) == 0 and len(arr_indices) != 0:
                raise IndexError(non_empty_take_msg)

            result = self[arr_indices]

        return result

    @classmethod
    def _concat_same_type(
        cls,
        to_concat: Sequence[T],
    ) -> T:
        """
        Concatenate multiple array.

        Parameters:
            to_concat: Sequence of FData objects to concat.

        Returns:
            Concatenation of the objects.

        """
        if isinstance(to_concat, cls):
            return to_concat

        return concatenate(to_concat)

    def astype(self, dtype: Any, copy: bool = True) -> Any:
        """Cast to a new dtype."""
        if isinstance(dtype, type(self.dtype)):
            new_obj = self
            if copy:
                new_obj = self.copy()
            return new_obj

        return super().astype(dtype)

    def _reduce(self, name: str, skipna: bool = True, **kwargs: Any) -> Any:
        meth = getattr(self, name, None)
        if meth:
            return meth(skipna=skipna, **kwargs)

        raise TypeError(
            f"'{type(self).__name__}' does not implement "
            f"reduction '{name}'",
        )


def concatenate(functions: Iterable[T], as_coordinates: bool = False) -> T:
    """
    Join samples from an iterable of similar FData objects.

    Joins samples of FData objects if they have the same
    dimensions and sampling points.
    Args:
        objects: Objects to be concatenated.
        as_coordinates:  If False concatenates as
                new samples, else, concatenates the other functions as
                new components of the image. Defaults to False.
    Returns:
        FData object with the samples from the
        original objects.
    Raises:
        ValueError: In case the provided list of FData objects is
        empty.
    Todo:
        By the moment, only unidimensional objects are supported in basis
        representation.
    """
    functions = iter(functions)
    first = next(functions, None)

    if first is None:
        raise ValueError(
            "At least one FData object must be provided to concatenate.",
        )

    return first.concatenate(*functions, as_coordinates=as_coordinates)


F = TypeVar("F", covariant=True)


class _CoordinateSequence(Protocol[F]):
    """
    Sequence of coordinates.

    Note that this represents a sequence of coordinates, not a sequence of
    FData objects.
    """

    def __getitem__(
        self,
        key: Union[int, slice],
    ) -> F:
        pass

    def __len__(self) -> int:
        pass



"""Test the covariance method of FData."""
from typing import Any, Callable, Tuple

import numpy as np
import pytest

from skfda.datasets import make_gaussian_process
from skfda.misc.covariances import CovarianceLike, Gaussian
from skfda.representation import FDataBasis
from skfda.representation.basis import Basis, BSplineBasis, FourierBasis

############
# FIXTURES
############

N_SAMPLES = 100
N_FEATURES = 100


@pytest.fixture(
    params=[
        FourierBasis,
        BSplineBasis,
    ],
    ids=[
        "FourierBasis",
        "BSplineBasis",
    ],
)
def basis_type(
    request: Any,
) -> Any:
    """Fixture for classes to test."""
    return request.param


@pytest.fixture(
    params=[
        5,
        30,
    ],
)
def n_basis(
    request: Any,
) -> Any:
    """Generate a basis."""
    return request.param


@pytest.fixture(
    params=[
        (0.0, 1.0),
        (-2.0, 2.0),
    ],
)
def interval(
    request: Any,
) -> Any:
    """Generate an interval."""
    return request.param


@pytest.fixture(
    params=[
        Gaussian(
            variance=1,
            length_scale=1,
        ),
        Gaussian(
            variance=0.1,
            length_scale=1,
        ),
        Gaussian(
            variance=3,
            length_scale=0.5,
        ),
    ],
)
def covariance(
    request: Any,
) -> Any:
    """Generate a covariance kernel."""
    return request.param


@pytest.fixture
def data_in_basis(
    basis_type: Callable[..., Basis],
    n_basis: int,
    interval: Tuple[float, float],
    covariance: CovarianceLike,
) -> FDataBasis:
    """Generate gaussian process data using a basis."""
    basis = basis_type(
        n_basis=n_basis,
        domain_range=interval,
    )
    return make_gaussian_process(
        n_samples=N_SAMPLES,
        n_features=N_FEATURES,
        start=interval[0],
        stop=interval[1],
        cov=covariance,
    ).to_basis(basis)


############
# TESTS
############

def test_overload_cov() -> None:
    """Test that the overloading of the cov method is consistent."""
    # Generate any sample data
    data = make_gaussian_process(
        n_samples=N_SAMPLES,
        n_features=N_FEATURES,
        start=0,
        stop=1,
    )
    grid_points = np.linspace(0, 1, N_FEATURES)
    # Test for FDataGrid
    np.testing.assert_equal(
        data.cov()(grid_points, grid_points),
        data.cov(grid_points, grid_points),
    )
    # Test for FDataBasis
    basis = FourierBasis(n_basis=5, domain_range=(0, 1))
    data_in_basis = data.to_basis(basis)
    np.testing.assert_equal(
        data_in_basis.cov()(grid_points, grid_points),
        data_in_basis.cov(grid_points, grid_points),
    )


def test_fdatabasis_covariance(
    data_in_basis: FDataBasis,
) -> None:
    """Test the covariance method of FDataBasis.

    The resulting covariance function is defined on the tensor basis that is
    the tensor product of the basis of the original functions with itself.

    This test checks that the covariance function calculated for FDataBasis
    objects is equal to the covariance function calculated for the grid
    representation of the same data.
    """
    # Select grid points before converting to grid
    domain_range = data_in_basis.domain_range[0]
    grid_points = np.linspace(domain_range[0], domain_range[1], N_FEATURES)
    data_in_grid = data_in_basis.to_grid(grid_points)
    # Check that the covariance functions are equal
    np.testing.assert_allclose(
        data_in_basis.cov(grid_points, grid_points),
        data_in_grid.cov(grid_points, grid_points),
    )



from typing import Any, TypeVar

import numpy as np
from scipy.interpolate import InterpolatedUnivariateSpline
from scipy.interpolate.interpnd import LinearNDInterpolator

from ..._utils._sklearn_adapter import BaseEstimator, InductiveTransformerMixin
from ...representation import FDataGrid
from ...typing._base import GridPoints
from ...typing._numpy import NDArrayFloat, NDArrayInt

T = TypeVar("T", bound=FDataGrid)


def _coords_from_indices(
    coord_indices: NDArrayInt,
    grid_points: GridPoints,
) -> NDArrayFloat:
    return np.stack([
        grid_points[i][coord_index]
        for i, coord_index in enumerate(coord_indices.T)
    ]).T


def _interpolate_nans(
    fdatagrid: T,
) -> T:

    data_matrix = fdatagrid.data_matrix.copy()

    for n_sample in range(fdatagrid.n_samples):
        for n_coord in range(fdatagrid.dim_codomain):

            data_points = data_matrix[n_sample, ..., n_coord]
            nan_pos = np.isnan(data_points)
            valid_pos = ~nan_pos
            coord_indices = np.argwhere(valid_pos)
            desired_coord_indices = np.argwhere(nan_pos)
            coords = _coords_from_indices(
                coord_indices,
                fdatagrid.grid_points,
            )
            desired_coords = _coords_from_indices(
                desired_coord_indices,
                fdatagrid.grid_points,
            )
            values = data_points[valid_pos]

            if fdatagrid.dim_domain == 1:
                interpolation = InterpolatedUnivariateSpline(
                    coords,
                    values,
                    k=1,
                    ext=3,
                )
            else:
                interpolation = LinearNDInterpolator(
                    coords,
                    values,
                )

            new_values = interpolation(
                desired_coords,
            )

            data_matrix[n_sample, nan_pos, n_coord] = new_values.ravel()

    return fdatagrid.copy(data_matrix=data_matrix)


class MissingValuesInterpolation(
    BaseEstimator,
    InductiveTransformerMixin[T, T, Any],
):
    """
    Class to interpolate missing values.

    Missing values are represented as NaNs.
    They are interpolated from nearby values with valid data.
    Note that this may be a poor choice if there are large contiguous portions
    of the function with missing values, as some of them would be inferred from
    very far away points.

    Examples:
        It is possible to interpolate NaNs scalar-valued univariate functions:

        >>> from skfda import FDataGrid
        >>> from skfda.preprocessing.missing import MissingValuesInterpolation
        >>> import numpy as np

        >>> X = FDataGrid([
        ...     [1, 2, np.nan, 4],
        ...     [5, np.nan, 7, 8],
        ...     [9, 10, np.nan, 12],
        ... ])
        >>> nan_interp = MissingValuesInterpolation()
        >>> X_transformed = nan_interp.fit_transform(X)
        >>> X_transformed.data_matrix[..., 0]
        array([[ 1.,  2.,  3.,  4.],
               [ 5.,  6.,  7.,  8.],
               [ 9., 10., 11., 12.]])

        For vector-valued functions each coordinate is interpolated
        independently:

        >>> X = FDataGrid(
        ...     [
        ...         [
        ...             (1, 5),
        ...             (2, np.nan),
        ...             (np.nan, 7),
        ...             (4, 8),
        ...         ],
        ...         [
        ...             (9, 13),
        ...             (10, np.nan),
        ...             (np.nan, np.nan),
        ...             (12, 16),
        ...         ],
        ...     ],
        ...     grid_points=np.linspace(0, 1, 4)
        ... )
        >>> nan_interp = MissingValuesInterpolation()
        >>> X_transformed = nan_interp.fit_transform(X)
        >>> X_transformed.data_matrix # doctest: +NORMALIZE_WHITESPACE
        array([[[  1.,  5.],
                [  2.,  6.],
                [  3.,  7.],
                [  4.,  8.]],
               [[  9., 13.],
                [ 10., 14.],
                [ 11., 15.],
                [ 12., 16.]]])

        For multivariate functions, such as surfaces all dimensions are
        considered. This is currently done using
        :external:class:`~scipy.interpolate.LinearNDInterpolator`, which
        triangulates the space and performs linear barycentric interpolation:

        >>> X = FDataGrid(
        ...     [
        ...         [
        ...             [1, 2, 3, 4],
        ...             [5, np.nan, 7, 8],
        ...             [10, 10, np.nan, 10],
        ...             [13, 14, 15, 16],
        ...         ],
        ...     ],
        ...     grid_points=(np.linspace(0, 1, 4), np.linspace(0, 1, 4))
        ... )
        >>> nan_interp = MissingValuesInterpolation()
        >>> X_transformed = nan_interp.fit_transform(X)
        >>> X_transformed.data_matrix[..., 0]
        array([[[  1.,   2.,   3.,   4.],
                [  5.,   6.,   7.,   8.],
                [ 10.,  10.,  11.,  10.],
                [ 13.,  14.,  15.,  16.]]])
    """

    def transform(
        self,
        X: T,
    ) -> T:
        return _interpolate_nans(X)



"""Regression."""
from typing import TYPE_CHECKING

import lazy_loader as lazy

__getattr__, __dir__, __all__ = lazy.attach(
    __name__,
    submod_attrs={
        "_historical_linear_model": ["HistoricalLinearRegression"],
        "_kernel_regression": ["KernelRegression"],
        "_linear_regression": ["LinearRegression"],
        "_fpca_regression": ["FPCARegression"],
        "_neighbors_regression": [
            "KNeighborsRegressor",
            "RadiusNeighborsRegressor",
        ],
    },
)

if TYPE_CHECKING:
    from ._fpca_regression import FPCARegression
    from ._historical_linear_model import (
        HistoricalLinearRegression as HistoricalLinearRegression,
    )
    from ._kernel_regression import KernelRegression as KernelRegression
    from ._linear_regression import LinearRegression as LinearRegression
    from ._neighbors_regression import (
        KNeighborsRegressor as KNeighborsRegressor,
        RadiusNeighborsRegressor as RadiusNeighborsRegressor,
    )



from typing import TYPE_CHECKING

import lazy_loader as lazy

__getattr__, __dir__, __all__ = lazy.attach(
    __name__,
    submod_attrs={
        "_regularization": [
            "L2Regularization",
            "TikhonovRegularization",
            "compute_penalty_matrix",
        ],

    },
)

if TYPE_CHECKING:
    from ._regularization import (
        L2Regularization as L2Regularization,
        TikhonovRegularization as TikhonovRegularization,
        compute_penalty_matrix as compute_penalty_matrix,
    )



"""Tests for outlying measures."""
import unittest

import numpy as np

from skfda import FDataGrid
from skfda.exploratory.depth.multivariate import SimplicialDepth
from skfda.exploratory.outliers import (
    MSPlotOutlierDetector,
    directional_outlyingness_stats,
)


class TestsDirectionalOutlyingness(unittest.TestCase):
    """Tests for directional outlyingness."""

    def test_directional_outlyingness(self) -> None:
        """Test non-asymptotic behaviour."""
        data_matrix = [
            [[0.3], [0.4], [0.5], [0.6]],
            [[0.5], [0.6], [0.7], [0.7]],
            [[0.2], [0.3], [0.4], [0.5]],
        ]
        grid_points = [2, 4, 6, 8]
        fd = FDataGrid(data_matrix, grid_points)
        stats = directional_outlyingness_stats(
            fd,
            multivariate_depth=SimplicialDepth(),
        )
        np.testing.assert_allclose(
            stats.directional_outlyingness,
            np.array([
                np.tile(0.0, (4, 1)),
                np.tile(0.5, (4, 1)),
                np.tile(-0.5, (4, 1)),
            ]),
            rtol=1e-06,
        )
        np.testing.assert_allclose(
            stats.mean_directional_outlyingness,
            np.array([
                [0.0],
                [0.5],
                [-0.5],
            ]),
            rtol=1e-06,
        )
        np.testing.assert_allclose(
            stats.variation_directional_outlyingness,
            np.array([0, 0, 0]),
            atol=1e-6,
        )

    def test_asymptotic_formula(self) -> None:
        """Test the asymptotic behaviour."""
        data_matrix = [
            [1, 1, 2, 3, 2.5, 2],
            [0.5, 0.5, 1, 2, 1.5, 1],
            [-1, -1, -0.5, 1, 1, 0.5],  # noqa: WPS204
            [-0.5, -0.5, -0.5, -1, -1, -1],
        ]
        grid_points = [0, 2, 4, 6, 8, 10]
        fd = FDataGrid(data_matrix, grid_points)
        out_detector = MSPlotOutlierDetector(
            _force_asymptotic=True,
        )
        prediction = out_detector.fit_predict(fd)
        np.testing.assert_allclose(
            prediction,
            np.array([1, 1, 1, 1]),
        )


if __name__ == '__main__':
    unittest.main()



from typing import TYPE_CHECKING

import lazy_loader as lazy

__getattr__, __dir__, __all__ = lazy.attach(
    __name__,
    submod_attrs={
        "_functions": [
            "average_function_value",
        ],
    },
)

if TYPE_CHECKING:

    from ._functions import (
        average_function_value as average_function_value,
    )



"""Test FDataGrid behaviour."""
import unittest
from typing import Sequence, Tuple

import numpy as np
import scipy.stats.mstats
from mpl_toolkits.mplot3d import axes3d

from skfda import FDataGrid, concatenate
from skfda.exploratory import stats


class TestFDataGrid(unittest.TestCase):
    """Test the FDataGrid representation."""

    def test_init(self) -> None:
        """Test creation."""
        fd = FDataGrid([[1, 2, 3, 4, 5], [2, 3, 4, 5, 6]])
        np.testing.assert_array_equal(
            fd.data_matrix[..., 0],
            np.array([[1, 2, 3, 4, 5], [2, 3, 4, 5, 6]]),
        )
        np.testing.assert_array_equal(fd.sample_range, [(0, 1)])
        np.testing.assert_array_equal(
            fd.grid_points, np.array([[0, 0.25, 0.5, 0.75, 1]]),
        )

    def test_copy_equals(self) -> None:
        """Test that copies compare equals."""
        fd = FDataGrid([[1, 2, 3, 4, 5], [2, 3, 4, 5, 6]])
        self.assertTrue(fd.equals(fd.copy()))

    def test_mean(self) -> None:
        """Test aritmetic mean."""
        fd = FDataGrid([[1, 2, 3, 4, 5], [2, 3, 4, 5, 6]])
        mean = stats.mean(fd)
        np.testing.assert_array_equal(
            mean.data_matrix[0, ..., 0],
            np.array([1.5, 2.5, 3.5, 4.5, 5.5]),
        )
        np.testing.assert_array_equal(fd.sample_range, [(0, 1)])
        np.testing.assert_array_equal(
            fd.grid_points,
            np.array([[0, 0.25, 0.5, 0.75, 1]]),
        )

    def test_gmean(self) -> None:
        """Test geometric mean."""
        fd = FDataGrid([[1, 2, 3, 4, 5], [2, 3, 4, 5, 6]])
        mean = stats.gmean(fd)
        np.testing.assert_array_equal(
            mean.data_matrix[0, ..., 0],
            scipy.stats.mstats.gmean(
                np.array([[1, 2, 3, 4, 5], [2, 3, 4, 5, 6]]),
            ),
        )
        np.testing.assert_array_equal(fd.sample_range, [(0, 1)])
        np.testing.assert_array_equal(
            fd.grid_points,
            np.array([[0, 0.25, 0.5, 0.75, 1]]),
        )

    def test_slice(self) -> None:
        """Test slicing behaviour."""
        t = (5, 3)
        fd = FDataGrid(data_matrix=np.ones(t))
        fd = fd[1:3]
        np.testing.assert_array_equal(
            fd.data_matrix[..., 0],
            np.array([[1, 1, 1], [1, 1, 1]]),
        )

    def test_concatenate(self) -> None:
        """
        Test concatenation.

        Ensure that the original argument and coordinate names are kept.

        """
        fd1 = FDataGrid([[1, 2, 3, 4, 5], [2, 3, 4, 5, 6]])
        fd2 = FDataGrid([[3, 4, 5, 6, 7], [4, 5, 6, 7, 8]])

        fd1.argument_names = ("x",)
        fd1.coordinate_names = ("y",)
        fd = fd1.concatenate(fd2)

        self.assertEqual(fd.n_samples, 4)
        self.assertEqual(fd.dim_codomain, 1)
        self.assertEqual(fd.dim_domain, 1)
        np.testing.assert_array_equal(
            fd.data_matrix[..., 0],
            [
                [1, 2, 3, 4, 5],
                [2, 3, 4, 5, 6],
                [3, 4, 5, 6, 7],
                [4, 5, 6, 7, 8],
            ],
        )
        self.assertEqual(
            fd1.argument_names,
            fd.argument_names,
        )
        self.assertEqual(
            fd1.coordinate_names,
            fd.coordinate_names,
        )

    def test_concatenate_coordinates(self) -> None:
        """
        Test concatenation as coordinates.

        Ensure that the coordinate names are concatenated.

        """
        fd1 = FDataGrid([[1, 2, 3, 4], [2, 3, 4, 5]])
        fd2 = FDataGrid([[3, 4, 5, 6], [4, 5, 6, 7]])

        fd1.argument_names = ("x",)
        fd1.coordinate_names = ("y",)
        fd2.argument_names = ("w",)
        fd2.coordinate_names = ("t",)
        fd = fd1.concatenate(fd2, as_coordinates=True)

        self.assertEqual(fd.n_samples, 2)
        self.assertEqual(fd.dim_codomain, 2)
        self.assertEqual(fd.dim_domain, 1)

        np.testing.assert_array_equal(
            fd.data_matrix,
            [
                [[1, 3], [2, 4], [3, 5], [4, 6]],
                [[2, 4], [3, 5], [4, 6], [5, 7]],
            ],
        )

        # Testing labels
        self.assertEqual(("y", "t"), fd.coordinate_names)
        fd2.coordinate_names = None  # type: ignore[assignment]
        fd = fd1.concatenate(fd2, as_coordinates=True)
        self.assertEqual(("y", None), fd.coordinate_names)
        fd1.coordinate_names = None  # type: ignore[assignment]
        fd = fd1.concatenate(fd2, as_coordinates=True)
        self.assertEqual((None, None), fd.coordinate_names)

    def test_concatenate_function(self) -> None:
        """Test the concatenate function (as opposed to method)."""
        sample1 = np.arange(0, 10)
        sample2 = np.arange(10, 20)
        fd1 = FDataGrid([sample1])
        fd2 = FDataGrid([sample2])

        fd1.argument_names = ("x",)
        fd1.coordinate_names = ("y",)
        fd = concatenate([fd1, fd2])

        self.assertEqual(fd.n_samples, 2)
        self.assertEqual(fd.dim_codomain, 1)
        self.assertEqual(fd.dim_domain, 1)
        np.testing.assert_array_equal(
            fd.data_matrix[..., 0],
            [sample1, sample2],
        )
        self.assertEqual(
            fd1.argument_names,
            fd.argument_names,
        )
        self.assertEqual(
            fd1.coordinate_names,
            fd.coordinate_names,
        )

    def test_coordinates(self) -> None:
        """Test coordinate access and iteration."""
        fd1 = FDataGrid([[1, 2, 3, 4], [2, 3, 4, 5]])
        fd1.argument_names = ("x",)
        fd1.coordinate_names = ("y",)
        fd2 = FDataGrid([[3, 4, 5, 6], [4, 5, 6, 7]])
        fd = fd1.concatenate(fd2, as_coordinates=True)

        # Indexing with number
        np.testing.assert_array_equal(
            fd.coordinates[0].data_matrix,
            fd1.data_matrix,
        )
        np.testing.assert_array_equal(
            fd.coordinates[1].data_matrix,
            fd2.data_matrix,
        )

        # Iteration
        for fd_j, fd_i in zip([fd1, fd2], fd.coordinates):
            np.testing.assert_array_equal(fd_j.data_matrix, fd_i.data_matrix)

        fd3 = fd1.concatenate(fd2, fd1, fd, as_coordinates=True)

        # Multiple indexation
        self.assertEqual(fd3.dim_codomain, 5)
        np.testing.assert_array_equal(
            fd3.coordinates[:2].data_matrix,
            fd.data_matrix,
        )
        np.testing.assert_array_equal(
            fd3.coordinates[-2:].data_matrix,
            fd.data_matrix,
        )
        index_original = np.array((False, False, True, False, True))
        np.testing.assert_array_equal(
            fd3.coordinates[index_original].data_matrix,
            fd.data_matrix,
        )

    def test_add(self) -> None:
        """Test addition with different objects."""
        fd1 = FDataGrid([[1, 2, 3, 4], [2, 3, 4, 5]])

        fd2 = fd1 + fd1
        np.testing.assert_array_equal(
            fd2.data_matrix[..., 0],  # noqa: WPS204
            [[2, 4, 6, 8], [4, 6, 8, 10]],
        )

        fd2 = fd1 + 2
        np.testing.assert_array_equal(
            fd2.data_matrix[..., 0],
            [[3, 4, 5, 6], [4, 5, 6, 7]],
        )

        fd2 = fd1 + np.array(2)
        np.testing.assert_array_equal(
            fd2.data_matrix[..., 0],
            [[3, 4, 5, 6], [4, 5, 6, 7]],
        )

        fd2 = fd1 + np.array([2])
        np.testing.assert_array_equal(
            fd2.data_matrix[..., 0],
            [[3, 4, 5, 6], [4, 5, 6, 7]],
        )

        fd2 = fd1 + np.array([1, 2])
        np.testing.assert_array_equal(
            fd2.data_matrix[..., 0],
            [[2, 3, 4, 5], [4, 5, 6, 7]],
        )

    def test_composition(self) -> None:
        """Test function composition."""
        X, Y, Z = axes3d.get_test_data(1.2)

        data_matrix = [Z.T]
        grid_points = [X[0, :], Y[:, 0]]

        g = FDataGrid(data_matrix, grid_points)
        self.assertEqual(g.dim_domain, 2)
        self.assertEqual(g.dim_codomain, 1)

        t = np.linspace(0, 2 * np.pi, 100)

        data_matrix = [10 * np.array([np.cos(t), np.sin(t)]).T]
        f = FDataGrid(data_matrix, t)
        self.assertEqual(f.dim_domain, 1)
        self.assertEqual(f.dim_codomain, 2)

        gof = g.compose(f)
        self.assertEqual(gof.dim_domain, 1)
        self.assertEqual(gof.dim_codomain, 1)


class TestEvaluateFDataGrid(unittest.TestCase):
    """Test FDataGrid evaluation."""

    def setUp(self) -> None:
        """Create bidimensional FDataGrid."""
        data_matrix = np.array(
            [
                np.tile([0, 1, 2], (2, 2, 1)),
                np.tile([3, 4, 5], (2, 2, 1)),
            ])

        grid_points = [[0, 1], [0, 1]]

        fd = FDataGrid(data_matrix, grid_points=grid_points)
        self.assertEqual(fd.n_samples, 2)
        self.assertEqual(fd.dim_domain, 2)
        self.assertEqual(fd.dim_codomain, 3)

        self.fd = fd

    def test_evaluate_aligned(self) -> None:
        """Check normal evaluation."""
        res = self.fd([(0, 0), (1, 1), (2, 2), (3, 3)])
        expected = np.array([
            np.tile([0, 1, 2], (4, 1)),
            np.tile([3, 4, 5], (4, 1)),
        ])

        np.testing.assert_allclose(res, expected)

    def test_evaluate_unaligned(self) -> None:
        """Check unaligned evaluation."""
        res = self.fd(
            [
                [(0, 0), (1, 1), (2, 2), (3, 3)],
                [(1, 7), (5, 2), (3, 4), (6, 1)],
            ],
            aligned=False,
        )
        expected = np.array(
            [
                np.tile([0, 1, 2], (4, 1)),
                np.tile([3, 4, 5], (4, 1)),
            ],
        )

        np.testing.assert_allclose(res, expected)

    def test_evaluate_grid_aligned(self) -> None:
        """Test evaluation in aligned grid."""
        res = self.fd([[0, 1], [1, 2]], grid=True)
        expected = np.array([
            np.tile([0, 1, 2], (2, 2, 1)),
            np.tile([3, 4, 5], (2, 2, 1)),
        ])

        np.testing.assert_allclose(res, expected)

    def test_evaluate_grid_unaligned(self) -> None:
        """Test evaluation with a different grid per curve."""
        res = self.fd(
            [[[0, 1], [1, 2]], [[3, 4], [5, 6]]],
            grid=True,
            aligned=False,
        )
        expected = np.array([
            np.tile([0, 1, 2], (2, 2, 1)),
            np.tile([3, 4, 5], (2, 2, 1)),
        ])

        np.testing.assert_allclose(res, expected)

    def test_restrict(self) -> None:
        """Test FDataGrid.restrict with bounds."""
        # Test 1 sample function R^3 -> R^5.
        grid_points = ([0, 1], [0, 1, 2], [0, 1, 2, 3])
        data_matrix = np.ones((1, 2, 3, 4, 5))
        fd = FDataGrid(data_matrix, grid_points)
        restricted_domain = ((0, 1), (0.5, 1.5), (0.5, 2))
        fd_restricted = fd.restrict(restricted_domain, with_bounds=True)
        res = fd_restricted.grid_points
        expected: Tuple[Sequence[float], ...] = (
            [0, 1],
            [0.5, 1, 1.5],
            [0.5, 1, 2],
        )
        for r, e in zip(res, expected):
            np.testing.assert_array_equal(r, e)


if __name__ == '__main__':
    unittest.main()



"""Feature extraction."""
from typing import TYPE_CHECKING

import lazy_loader as lazy

__getattr__, __dir__, __all__ = lazy.attach(
    __name__,
    submod_attrs={
        "_coefficients_transformer": ["CoefficientsTransformer"],
        "_evaluation_trasformer": ["EvaluationTransformer"],
        "_fda_feature_union": ["FDAFeatureUnion"],
        "_function_transformers": [
            "LocalAveragesTransformer",
            "NumberCrossingsTransformer",
            "OccupationMeasureTransformer",
        ],
        "_functions": [
            "local_averages",
            "number_crossings",
            "occupation_measure",
            "unconditional_central_moment",
            "unconditional_expected_value",
            "unconditional_moment",
        ],
        "_per_class_transformer": ["PerClassTransformer"],
    },
)

if TYPE_CHECKING:
    from ._coefficients_transformer import (
        CoefficientsTransformer as CoefficientsTransformer,
    )
    from ._evaluation_trasformer import (
        EvaluationTransformer as EvaluationTransformer,
    )
    from ._fda_feature_union import FDAFeatureUnion as FDAFeatureUnion
    from ._function_transformers import (
        LocalAveragesTransformer as LocalAveragesTransformer,
        NumberCrossingsTransformer as NumberCrossingsTransformer,
        OccupationMeasureTransformer as OccupationMeasureTransformer,
    )
    from ._functions import (
        local_averages as local_averages,
        number_crossings as number_crossings,
        occupation_measure as occupation_measure,
        unconditional_central_moment as unconditional_central_moment,
        unconditional_expected_value as unconditional_expected_value,
        unconditional_moment as unconditional_moment,
    )
    from ._per_class_transformer import (
        PerClassTransformer as PerClassTransformer,
    )



"""Clustering."""

from typing import TYPE_CHECKING

import lazy_loader as lazy

__getattr__, __dir__, __all__ = lazy.attach(
    __name__,
    submod_attrs={
        "_hierarchical": ["AgglomerativeClustering"],
        "_kmeans": ["FuzzyCMeans", "KMeans"],
        "_neighbors_clustering": ["NearestNeighbors"],
    },
)

if TYPE_CHECKING:
    from ._hierarchical import (
        AgglomerativeClustering as AgglomerativeClustering,
    )
    from ._kmeans import FuzzyCMeans as FuzzyCMeans, KMeans as KMeans
    from ._neighbors_clustering import NearestNeighbors as NearestNeighbors



"""scikit-fda package."""
import errno as _errno
import os as _os
from typing import TYPE_CHECKING

import lazy_loader as lazy

__getattr__, __dir__, __all__ = lazy.attach(
    __name__,
    submodules=[
        "datasets",
        "exploratory",
        "inference",
        "misc",
        "ml",
        "preprocessing",
        "representation",
    ],
    submod_attrs={
        'representation': ["FData", "FDataBasis", "FDataGrid"],
        'representation._functional_data': ['concatenate'],
    },
)

if TYPE_CHECKING:
    from .representation import (
        FData as FData,
        FDataBasis as FDataBasis,
        FDataGrid as FDataGrid,
        concatenate as concatenate,
    )

__version__ = "0.9"



"""Classification."""

from typing import TYPE_CHECKING

import lazy_loader as lazy

__getattr__, __dir__, __all__ = lazy.attach(
    __name__,
    submod_attrs={
        "_centroid_classifiers": [
            "DTMClassifier",
            "NearestCentroid",
        ],
        "_depth_classifiers": [
            "DDClassifier",
            "DDGClassifier",
            "MaximumDepthClassifier",
        ],
        "_logistic_regression": ["LogisticRegression"],
        "_neighbors_classifiers": [
            "KNeighborsClassifier",
            "RadiusNeighborsClassifier",
        ],
        "_qda": ["QuadraticDiscriminantAnalysis"],
    },
)

if TYPE_CHECKING:
    from ._centroid_classifiers import (
        DTMClassifier as DTMClassifier,
        NearestCentroid as NearestCentroid,
    )
    from ._depth_classifiers import (
        DDClassifier as DDClassifier,
        DDGClassifier as DDGClassifier,
        MaximumDepthClassifier as MaximumDepthClassifier,
    )
    from ._logistic_regression import LogisticRegression as LogisticRegression
    from ._neighbors_classifiers import (
        KNeighborsClassifier as KNeighborsClassifier,
        RadiusNeighborsClassifier as RadiusNeighborsClassifier,
    )
    from ._qda import (
        QuadraticDiscriminantAnalysis as QuadraticDiscriminantAnalysis,
    )
